# run with : python3 -m unittest TestKepplers2ndLaw

import unittest
import os
import numpy as numpy
import math
from astropy import units
import matplotlib.pyplot as plt
import warnings
from numpy.testing import assert_array_almost_equal as assertAAE

from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2019
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestKepplers2ndLaw( unittest.TestCase ):
    """
    Test harness for Kepplers 2nd law

    Author:      Do Kester

    """
    TWOPI = 2 * math.pi

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def test1( self ):
        x  = numpy.linspace( 0.0, 10, 101, dtype=float )
        print( "******KEPPLERS LAW***************" )
        KL = Kepplers2ndLaw( )

        # eccen, semimajor, period, periastron
        p = [0.0, 1.0, 10.0, 0.0]
        m = KL.meanAnomaly( x, p )
        e = KL.eccentricAnomaly( x, p )
        r, v = KL.radiusAndTrueAnomaly( x, p )
        v = numpy.where( v < 0, v + self.TWOPI, v )
        print( "Mean ", fmt( m ) )
        print( "Ecce ", fmt( e ) )
        print( "True ", fmt( v ) )
        print( "Rad  ", fmt( r ) )

        se = KL.sinE
        ce = KL.cosE

        for k in range( 101 ) :
            print( fmt(k), fmt(m[k]), fmt(e[k]), fmt(v[k]), fmt(se[k]), fmt(ce[k]) )

        assertAAE( m, e )
        assertAAE( e, v )
        x0 = r * numpy.cos( v )
        y0 = r * numpy.sin( v )


        p = [0.5, 0.8, 1.4, 1.3]
        m = KL.meanAnomaly( x, p )
        e = KL.eccentricAnomaly( x, p )
        r, v = KL.radiusAndTrueAnomaly( x, p )
        v = numpy.where( v < 0, v + self.TWOPI, v )
        print( "Mean ", fmt( m ) )
        print( "Ecce ", fmt( e ) )
        print( "True ", fmt( v ) )
        print( "Rad  ", fmt( r ) )
        x1 = r * numpy.cos( v )
        y1 = r * numpy.sin( v )


        if self.doplot :
            plt.figure( 1, figsize=[6,6] )
            plt.plot( x0, y0, 'k-' )
            plt.plot( x1, y1, 'r-' )

            plt.show()

    def test2( self ):
        x  = numpy.linspace( 0, 10, 1001, dtype=float )
        print( "******KEPPLERS LAW test 2***************" )
        KL = Kepplers2ndLaw( )

        # eccen, semimajor, period, periastron
        p = [0.0, 1.0, 10.0, 0.0]
        e2 = None
        for k in range( 10 ) :
            e0 = KL.eccentricAnomaly0( x, p )
            i0 = KL.iter
            e1 = KL.eccentricAnomaly1( x, p )
            i1 = KL.iter
            e2 = KL.eccentricAnomaly2( x, p, Estart=e2 )
            i2 = KL.iter
            print( "Eccentricity  ", fmt(p[0]), fmt(i0), fmt(i1), fmt(i2) )
            p[0] += 0.1


        p[0] = 0.6
        e = KL.eccentricAnomaly( x, p, Estart=e2 )
        print( "Eccentricity  ", fmt(p[0]), "                   ", fmt( KL.iter ) )
        p[0] = 0.999999999
        e = KL.eccentricAnomaly( x, p, Estart=e2 )
        print( "Eccentricity  ", p[0], "                 ", fmt( KL.iter ) )

    def test3( self ):
        x  = numpy.linspace( 0, 10, 1001, dtype=float )
        print( "******KEPPLERS LAW test 3***************" )
        KL = Kepplers2ndLaw( )

        # eccen, semimajor, period, periastron
        p = [0.0, 2.0, 10.0, 0.0]

        for k in range( 10 ) :
            r, v = KL.radiusAndTrueAnomaly( x, p )

            a = 0.5 * ( r[0] + r[501] )
            b = math.sqrt( r[0] * r[501] )
            e = math.sqrt( a * a - b * b ) / a

            xx = r * numpy.cos( v )
            yy = r * numpy.sin( v )

            if self.doplot :
                plt.plot( xx, yy )

            print( fmt( p[0] ), fmt( e ), fmt( a ), fmt( b ), fmt( KL.iter ) )
            assertAAE( p[0], e )

            p[0] += 0.1

        if self.doplot :
            plt.plot( [0], [0], 'k.' )
            plt.show()


    def test4( self ):
        x  = numpy.linspace( 0, 10, 11, dtype=float )
        xp = x + 0.0001
        xm = x - 0.0001
        print( "******KEPPLERS LAW test 4***************" )
        KL = Kepplers2ndLaw( )

        print( "  Derivatives to x" )
        # eccen, semimajor, period, periastron
        p = [0.2, 2.0, 10.0, 0.3]

        mp = KL.meanAnomaly( xp, p )
        mm = KL.meanAnomaly( xm, p )
        nm = ( mp - mm ) / 0.0002
        dm = KL.dMdx( x, p )
        print( "dMdx  ", fma( dm ) )
        print( "Numer ", fma( nm ) )
        assertAAE( dm, nm )

        ep = KL.eccentricAnomaly( xp, p )
        em = KL.eccentricAnomaly( xm, p )
        ea = KL.eccentricAnomaly( x, p )
        cosE = numpy.cos( ea )
        ne = ( ep - em ) / 0.0002
        de = KL.dEdx( x, p, cosE )
        print( "dEdx  ", fma( de ) )
        print( "Numer ", fma( ne ) )
        assertAAE( de, ne )

        rp, vp = KL.radiusAndTrueAnomaly( xp, p )
        rm, vm = KL.radiusAndTrueAnomaly( xm, p )
        nr = ( rp - rm ) / 0.0002
        nv = ( vp - vm ) / 0.0002
        sinE = numpy.sin( ea )
        dr, dv = KL.drvdx( x, p, cosE, sinE )
        print( "drdx  ", fma( dr ) )
        print( "Numer ", fma( nr ) )
        assertAAE( dr, nr )
        print( "dvdx  ", fma( dv ) )
        print( "Numer ", fma( nv ) )
        assertAAE( dv, nv )


        print( "  Partials to parameters" )
        drde, drda, drdP, drdp, dvde, dvdP, dvdp = KL.drvdpar( x, p, ea, cosE, sinE )

        pp = p[:]
        pp[0] += 0.0001
        pm = p[:]
        pm[0] -= 0.0001

        rp, vp = KL.radiusAndTrueAnomaly( x, pp )
        rm, vm = KL.radiusAndTrueAnomaly( x, pm )
        nr = ( rp - rm ) / 0.0002
        nv = ( vp - vm ) / 0.0002

        print( "drde  ", fma( drde ) )
        print( "Numer ", fma( nr ) )
        assertAAE( drde, nr )
        print( "dvde  ", fma( dvde ) )
        print( "Numer ", fma( nv ) )
        assertAAE( dvde, nv )

        pp = p[:]
        pp[1] += 0.0001
        pm = p[:]
        pm[1] -= 0.0001

        rp, vp = KL.radiusAndTrueAnomaly( x, pp )
        rm, vm = KL.radiusAndTrueAnomaly( x, pm )
        nr = ( rp - rm ) / 0.0002
        nv = ( vp - vm ) / 0.0002

        print( "drda  ", fma( drda ) )
        print( "Numer ", fma( nr ) )
        assertAAE( drda, nr )
        dvda = numpy.zeros_like( drda )
        print( "dvda  ", fma( dvda ) )
        print( "Numer ", fma( nv ) )
        assertAAE( dvda, nv )

        pp = p[:]
        pp[2] += 0.0001
        pm = p[:]
        pm[2] -= 0.0001

        rp, vp = KL.radiusAndTrueAnomaly( x, pp )
        rm, vm = KL.radiusAndTrueAnomaly( x, pm )
        nr = ( rp - rm ) / 0.0002
        nv = ( vp - vm ) / 0.0002

        print( "drdP  ", fma( drdP ) )
        print( "Numer ", fma( nr ) )
        assertAAE( drdP, nr )
        print( "dvdP  ", fma( dvdP ) )
        print( "Numer ", fma( nv ) )
        assertAAE( dvdP, nv )

        pp = p[:]
        pp[3] += 0.0001
        pm = p[:]
        pm[3] -= 0.0001

        rp, vp = KL.radiusAndTrueAnomaly( x, pp )
        rm, vm = KL.radiusAndTrueAnomaly( x, pm )
        nr = ( rp - rm ) / 0.0002
        nv = ( vp - vm ) / 0.0002

        print( "drdp  ", fma( drdp ) )
        print( "Numer ", fma( nr ) )
        assertAAE( drdp, nr )
        print( "dvdp  ", fma( dvdp ) )
        print( "Numer ", fma( nv ) )
        assertAAE( dvdp, nv )



    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestKepplers2ndLaw.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestLMFitter

import unittest
import os
import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import math

import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import LevenbergMarquardtFitter as LMFitter

from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2004 - 2014 Do Kester, SRON (Java code)
#  *  2017        Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def makeData( self, bg=False ) :

        p = [3.2, -0.1, 0.3, 1.1, 2.1]
        ss = 0.3

        ndata = 101
        x = numpy.linspace( -1, +1, ndata, dtype=float )
        y = ( x - p[1] ) / p[2]
        numpy.random.seed( 3456 )
        y = p[0] * numpy.exp( -y * y ) + ss * numpy.random.randn( ndata )
        if not bg :
            return (x,y,p[:3])

        y += ( p[3] + p[4] * x )
        return (x,y,p)

    def eq( self, a, b, eps=1e-10 ) :
        if ( a + b ) != 0 :
            return abs( a - b ) / abs( a + b ) < eps
        else :
            return abs( a - b ) < eps


################################################################################
    def test1( self ):

        x,y,p1 = self.makeData()

        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing Nonlinear Fitters: LevenbergMarquardt (lm)" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        modl1 = GaussModel( )

        amfit = LMFitter( x, modl1 )

        print( "truth   ", fmt( p1 ) )
        par1 = amfit.fit( y )

        print( "pars1   ", fmt( par1 ) )
        assertAAE( par1, p1, 1 )
        print( "chisq1  ", amfit.chisq )
        print( "hessian \n", fmt( amfit.hessian ) )
        print( "design  \n", fmt( amfit.design[-4:,:], max=None ) )



        print( "With background" )
        x,z,p2 = self.makeData( bg=True )

        modl2 = GaussModel( )
        modl2.addModel( PolynomialModel(1) )

        modl2.parameters = numpy.append( par1, [0,0] )
        lmfit = LMFitter( x, modl2 )

        par2 = lmfit.fit( z )
        print( "truth   ", fmt( p2 ) )
        print( "pars2   ", fmt( par2 ) )

        print( "chisq2  ", lmfit.chisq )

        assertAAE( par2, p2, 1 )

        print( "hessian \n", fmt( lmfit.hessian ) )
        print( "design  \n", fmt( lmfit.design[-4:,:], max=None ) )

        if self.doplot :
            xx = numpy.linspace( -1, +1, 1001 )
            plt.plot( x, y, 'k+' )
            plt.plot( xx, modl1.result( xx ), 'k-' )
            plt.plot( x, z, 'r+' )
            plt.plot( xx, modl2.result( xx ), 'r-' )
            plt.show()


    def test2( self ):

        x,y,p1 = self.makeData()

        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing LevenbergMarquardt (normalized)" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        modl1 = GaussModel( )

        amfit = LMFitter( x, modl1 )

        par1 = amfit.fit( y )

        bmfit = LMFitter( x, modl1 )

        conpr1 = numpy.asarray( [1.0,0.0,0.0] )

        bmfit.normalize( conpr1, p1[0], weight=10.0 )

        print( "Normweight = ", bmfit.normweight )

        print( "truth  ", fmt( p1 ) )

        par3 = bmfit.fit( y )
        print( "pars3  ", fmt( par3 ) )
        print( "chisq  ", fmt( bmfit.chisq ) )

        hes1 = amfit.getHessian( params=par3 )
        print( "hessian\n", fmt( hes1 ) )
        hes3 = bmfit.getHessian( params=par3 )
        print( "hessian\n", fmt( hes3 ) )

        f = bmfit.normweight
        for h1,h3 in zip( hes1.flat, hes3.flat ) :
            self.assertTrue( abs( h1 + f - h3 ) < 1e-8 )
            f = 0.0

        if self.doplot :
            xx = numpy.linspace( -1, +1, 1001 )
            plt.plot( x, y, 'k+' )
            plt.plot( xx, modl1.result( xx ), 'g-' )
            plt.show()


    def test3( self ):

        x,y,p1 = self.makeData( bg=True )

        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing Linear Fitter with limits" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )


        print( "   1. no limits and keep in fit " )

        modl1 = GaussModel( )
        modl1.addModel( PolynomialModel(1) )

        amfit = LMFitter( x, modl1 )

        par0 = amfit.limitsFit( amfit.fit, y )
        print( "pars0   ", fmt( par0 ) )
        print( "stdv0   ", fmt( amfit.stdevs ) )
        print( "chisq0  ", amfit.chisq )

        par1 = amfit.limitsFit( amfit.fit, y, keep={0:3.3} )
        print( "pars1   ", fmt( par1 ) )
        print( "stdv1   ", fmt( amfit.stdevs ) )
        print( "chisq1  ", amfit.chisq )

        printclass( amfit )

        print( "   2. limits and keep in fit " )
        gm = GaussModel()
        gm.setLimits( [-3.0, -1.0, 0.1], [3.0, 1.0, 1.0] )
        pm = PolynomialModel( 1 )
        pm.setLimits( [-5], [5] )
        modl2 = gm + pm

        lmfit = LMFitter( x, modl2 )

        par2 = lmfit.limitsFit( lmfit.fit, y )
        print( "pars2   ", fmt( par2 ) )
        print( "stdv2   ", fmt( lmfit.stdevs ) )
        print( "chisq2  ", lmfit.chisq )

        par3 = lmfit.limitsFit( lmfit.fit, y, keep={0:3.3} )
        print( "pars3   ", fmt( par3 ) )
        print( "stdv3   ", fmt( lmfit.stdevs ) )
        print( "chisq3  ", lmfit.chisq )

        printclass( lmfit )

        print( "   2. limits and keep in fitter and fit " )
        gm = GaussModel()
        gm.setLimits( [-3.0, -1.0, 0.1], [3.0, 1.0, 1.0] )
        pm = PolynomialModel( 1 )
        pm.setLimits( [-5], [5] )
        modl3 = gm + pm

        lmfit = LMFitter( x, modl3, keep={4:5.0} )

        par4 = lmfit.limitsFit( lmfit.fit, y )
        print( "pars4   ", fmt( par4 ) )
        print( "stdv4   ", fmt( lmfit.stdevs ) )
        print( "chisq4  ", lmfit.chisq )

        par5 = lmfit.limitsFit( lmfit.fit, y, keep={0:3.3} )
        print( "pars5   ", fmt( par5 ) )
        print( "stdv5   ", fmt( lmfit.stdevs ) )
        print( "chisq5  ", lmfit.chisq )

        printclass( lmfit )

        if self.doplot :
            xx = numpy.linspace( -1, +1, 1001 )
            plt.plot( x, y, 'k+' )
            plt.plot( xx, modl1.result( xx ), 'k-' )

            plt.plot( xx, modl2.result( xx ), 'r-' )
            plt.show()

    def test5( self ):

        x,y,p1 = self.makeData( bg=True )

        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing Nonlinear Fitters: LevenbergMarquardt (lm)" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        modl1 = GaussModel( )
        modl1.addModel( PolynomialModel( 1 ) )

        amfit = LMFitter( x, modl1 )

        print( "truth   ", fmt( p1 ) )
        par1 = amfit.limitsFit( amfit.fit, y )

        print( "pars1   ", fmt( par1 ) )
        assertAAE( par1, p1, 1 )
        print( "chisq1  ", amfit.chisq )
        print( "hessian \n", fmt( amfit.hessian ) )
        print( "design  \n", fmt( amfit.design[-4:,:], max=None ) )


        gm = GaussModel( )
        gm.setLimits( [0.0, -1.0, 0.0], [3.2, 1.0, 1.0] )
        pm = PolynomialModel( 1 )
        pm.setLimits( [0.0, 0.0], [2.0, 2.0] )
        modl2 = gm + pm

        modl2.parameters = par1
        lmfit = LMFitter( x, modl2 )

        par2 = lmfit.limitsFit( lmfit.fit, y )
        print( "pars2   ", fmt( par2 ) )

        print( "chisq2  ", lmfit.chisq )

        assertAAE( par1, par2, 1 )

        print( "hessian \n", fmt( lmfit.hessian ) )
        print( "design  \n", fmt( lmfit.design[-4:,:], max=None ) )

        if self.doplot :
            xx = numpy.linspace( -1, +1, 1001 )
            plt.plot( x, y, 'k+' )
            plt.plot( xx, modl1.result( xx ), 'k-' )

            plt.plot( xx, modl2.result( xx ), 'r-' )
            plt.show()


    def test6( self ) :

        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing AmoebaFitter using a map" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )

        N = 71
        b0 = 3.5            # amplitude of source
        b1 = 30             # position
        b2 = 45             # position
        b3 = 13             # size of source
        b4 = 0.3            # background

        sig = 0.2

        x = numpy.arange( N, dtype=float )
        x1 = ( x - b1 ) / b3
        y1 = numpy.exp( - x1 * x1 )
        x2 = ( x - b2 ) / b3
        y2 = numpy.exp( - x2 * x2 )

        y = b0 * numpy.outer( y1, y2 ) + b4
        numpy.random.seed( 130105 )
        y += sig * numpy.random.randn( N, N )

        mdl = Kernel2dModel( kernel=Gauss(), shape='circular' )
        mdl += PolySurfaceModel( 0 )
        print( mdl )
        lo = [0, 0, 0, 0.01, 0 ]
        hi = [10, 70, 70, 70, 10 ]
        #mdl.setLimits( lo, hi )
        par = [5.0,30.0,30.0,30.0,5.0]
        mdl.parameters = par

        fitter = AmoebaFitter( y, mdl, map=True )
        print( "y      ", y.shape )
        print( "xdata  ", fitter.xdata.shape )
 
        # find the parameters
        param = fitter.fit( y, temp=10, maxiter=100000, size=5 )

        print( "yfit   ", fitter.yfit.shape )
        print( "mdlfit ", mdl( fitter.xdata ).shape )
        
        print( "Parameters :", fmt( param, max=None ) )
        print( "StDevs     :", fmt( fitter.stdevs, max=None ) )
        print( "Chisq      :", fmt( fitter.chisq ) )
        print( "Scale      :", fmt( fitter.scale ) )
        print( "Evidence   :", fmt( fitter.getEvidence( limits=[-100,100] ) ) )
        print( "Covar matrix :" )
        print( fmt( fitter.covariance))



    def test4( self ) :

        self.normalizetest( Fitter )
        self.normalizetest( QRFitter )
        self.normalizetest( CurveFitter )
        self.normalizetest( LMFitter )
        self.assertRaises( NotImplementedError, self.normalizetest, AmoebaFitter )
#        self.assertRaises( NotImplementedError, self.normalizetest, PowellFitter )

    def normalizetest( self, fitter ) :

        p = numpy.asarray( [2.0, 1.3] )
        x = numpy.asarray( [1.0, 1.3, 1.5, 1.8, 2.0] )
        numpy.random.seed( 12345 )
        y = p[0] + p[1] * x + 0.5 * numpy.random.randn( 5 )

        m = PolynomialModel( 1 )

        ftr = fitter( x, m )
        print( "=============================================================" )
        print( str( ftr ) )
        print( "xdata  ", fmt( x ) )
        print( "ydata  ", fmt( y ) )

        par = ftr.fit( y )
        print( "truth  ", fmt( p ) )
        print( "param  ", fmt( par ) )

        conpr = numpy.asarray( [1.0,0.0] )
        for w in [0,1,10,100,1000] :
            m2 = m.copy()
            ftr = fitter( x, m2 )

            ftr.normalize( conpr, p[0], weight=w )

            par = ftr.fit( y )
            print( fmt( w ), fmt( par ), fmt( ftr.chisq ) )

        self.assertTrue( abs( p[0] - par[0] ) < 1e-3 )

        print( fmt( ftr.hessian ) )
#        print( fmt( ftr.getVector( y ) ) )




    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestLMFitter.__class__ )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Sample import Sample

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2008 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class SampleList( list ):
    """
    SampleList is a list of Samples, see @Sample

    SampleList is the main result of the NestedSampler. It contains all
    information to calculate averages, medians, modi or maximum likihood solutions
    of the parameters, or of any function of the parameters; in particular of the
    Model.

    To make averages one has to take into account the weights. Each Sample has a weight
    and all weights sum to 1.0. So the average of any function, f, of the parameters p is

        E( f(p) ) = &sum; w_k f( p_k )

    where the sum is over all samples k.

    A large set of utility functions is provided to extract the information from the
    SampleList.


    Attributes
    ----------
    parameters : numpy.array (read-only)
        The average over the parameters. Not for dynamic models.
    stdevs, standardDeviations : numpy.array (read-only)
        The standard deviations for the parameters. Not for dynamic models
    scale : float
        The average of the noise scale
    stdevScale : float
        the standard deviation of the scale.

    logZ : float (read-only)
        Natural log of evidence
    evidence : float (read-only)
        log10( Z ). Evidence * 10 is interpretable as dB.
    info : float (read-only)
        The information H. The compression factor ( the ratio of the prior space
        available to the model parameters over the posterior space ) is equal to the exp( H ).

    maxLikelihoodIndex : int (read-only)
        The index at which the max likelihood can be found: always the last in the list
    maxLikelihoodParameters : numpy.array (read-only)
        The maximum likelihood parameters at the maxLikelihoodIndex.
    maxLikelihoodScale : float (read-only)
        The maximum likelihood noise scale at the maxLikelihoodIndex.
    medianIndex : int (read-only)
        The index at which the median can be found: the middle of the cumulative weights.
        It is calculated once and then kept.
    medianParameters : numpy.array (read-only)
        The median of the parameters at the medianIndex
    medianScale : float (read-only)
        The median of the noise scale at the medianIndex
    modusIndex : int (read-only)
        The index at which the modus can be found: the largest weight
        It is calculated once and then kept.
    modusParameters : numpy.array (read-only)
        The modus of the parameters at the modusIndex
    modusScale : float (read-only)
        The modus of the noise scale at the modusIndex.

    normalized : bool
        True when the weights are normalized to SUM( weights ) = 1


    Author       Do Kester

    """
    def __init__( self, model, nsamples, parameters=None, fitIndex=None, ndata=1 ):
        """
        Constructor.

        Parameters
        ----------
        nsamples : int
            number of samples created.
        model : Model
            to be used in the samples
        parameters : array_like
            list of model parameters
        fitIndex : array of int
            indicating which parameters need fitting
        ndata : int
            length of the data vector; to be used in stdev calculations

        """
        super( SampleList, self ).__init__( )
        self._count = 0
        self.iteration = 0
        self.logZ = 0.0
        self.info = 0.0
        self.addSamples( model, nsamples, parameters, fitIndex=fitIndex )
        self.maxLikelihoodIndex = -1            # always the last one
        self.normalized = False
        self.ndata = ndata

    def addSamples( self, model, nSamples, parameters, fitIndex=None ):
        for i in range( nSamples ) :
            if model.isDynamic() :
                model = model.copy()
                if fitIndex is not None :
                    fitIndex = fitIndex.copy()
            sample = Sample( self._count, -1, self.iteration, model,
                             parameters=parameters, fitIndex=fitIndex )
            self.append( sample )
            self._count += 1

    def __getattr__( self, name ) :
        if name == "parameters" :
            return self.getParameters()
        elif name == "stdevs" or name == "standardDeviations" :
            self.getParameters()
            return self.stdevs
        elif name == "hypars" :
            return self.getHypars()
        elif name == "stdevHypar" :
            self.getHypars()
            return self.stdevHypar
        elif name == "nuisance" :
            return self.getNuisance()
        elif name == "stdevNuis" :
            self.getNuisance()
            return self.stdevNuis
        elif name == "scale" :
            return self.hypars[0]
        elif name == "stdevScale" :
            return self.stdevHypars[0]

        elif name == "evidence" :
            return self.logZ / math.log( 10.0 )
        elif name == "maxLikelihoodParameters" :
            return self[self.maxLikelihoodIndex].parameters
        elif name == "maxLikelihoodScale" :
            return self[self.maxLikelihoodIndex].hypars[0]
        elif name == "medianIndex" :
            return self.getMedianIndex()
        elif name == "medianParameters" :
            return self[self.medianIndex].parameters
        elif name == "medianScale" :
            return self[self.medianIndex].hypars[0]
        elif name == "modusIndex" :
            we = list( self.getLogWeightEvolution() )
            self.modusIndex = we.index( max( we ) )
            return self.modusIndex
        elif name == "modusParameters" :
            return self[self.modusIndex].parameters
        elif name == "modusScale" :
            return self[self.modusIndex].hypars[0]


        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    # ===========================================================================
    def sample( self, k, sample=None ) :
        """
        Set or return the k-th sample from the list.

        Parameters
        ----------
        k : int
            the index of the sample
        sample : Sample
            if present, set the kth sample with sample
        """
        if sample is not None :
            self[k] = sample

        return self[k]


    def normalize( self ):
        """
        Normalize the samplelist.
        make Sum( weight ) = 1

        """
        self.normalized = True
        lwev = self.getLogWeightEvolution()

        lmax = numpy.max( lwev )
        lwev -= lmax

        lswt = math.log( numpy.sum( numpy.exp( lwev ) ) )

        for sample in self :
            sample.logW -= ( lmax + lswt )


    def add( self, sample ):
        """
        Add a Sample to the list

        Parameters
        ----------
        sample : Sample
            the sample to be added

        """
        sample.parent = sample.id       ## where it is from
        sample.id = self._count
        self._count += 1
        self.append( sample )
        self.normalized = False

    def copy( self, src, des ):
        """
        Copy one item of the list onto another.

        Parameters
        ----------
        src : int
            the source item
        des : int
            the destination item

        """
        id = self[des].id
        self[des] = self[src].copy()
        self[des].id = id


    def weed( self, maxsize=None ):
        """
        Weed superfluous samples.

        If MaxSamples has been set, it is checked whether the size of the
        SampleList exceeds the maximum. If so the Sample with the smallest
        log( Weight ) is removed.
        weed( ) is called recursively until the size has the required length.

        """
        if maxsize is None :
            return

        while len( self ) > maxsize :
            if self[0].logW < self[-1].logW :
                self.remove( self[0] )
            else :
                self.remove( self[-1] )
        return


    def logPlus( self, x, y ):
        """
        Return  log( exp(x) + exp(y) )
        """
        return numpy.logaddexp( x, y )

    # ===== AVERAGES and STDEVS ===================================================
    def getParameters( self ):
        """
        Calculate the average of the parameters and the standard deviations.

        Return
        ------
            The average values of the parameters.
        Raises
        ------
            ValueError when using Dynamic Models

        """
        mdl = self[0].model
        if not ( mdl.hasPriors() and any( [mdl.getPrior( k ).isCircular() for k in range( mdl.npchain )] ) ) :
            ( a, s ) = self.averstd( "parameters" )
            self.stdevs = s
            return a

        par, wgt = self.getParAndWgtEvolution()
        np = self[0].model.npchain
        param = numpy.zeros( np, dtype=float )
        stdev = numpy.zeros( np, dtype=float )
        for k in range( np ) :
            pr = mdl.getPrior( k )
            clh = None if not pr.isCircular() else pr.getLimits()
            param[k], stdev[k] = Tools.average( par[:,k], weights=wgt, circular=clh )

        self.parameters = param
        self.stdevs = stdev

        return self.parameters

        """
        np = self[0].model.npchain
        param = numpy.zeros( np, dtype=float )
        stdev = numpy.zeros( np, dtype=float )
        sumwt = 0.0
        for sample in self :
            if sample.model.npchain != np :
                raise ValueError( "Models with different " + "numbers of parameters: Cannot average" )
            wt = math.exp( sample.logW )
            sumwt += wt
            wp = wt * sample.parameters
            param = param + wp
            stdev = stdev + wp * sample.parameters

        stdev = numpy.sqrt( stdev - param * param )
        self.parameters = param
        self.stdevs = stdev

        return self.parameters
        """

    def getHypars( self ) :
        """
        Return the hyper parameters
        """
        nhp = len( self[0].hyper )

        hypar = numpy.zeros( nhp, dtype=float )
        hydev = numpy.zeros( nhp, dtype=float )
        sw = 0.0
        for sample in self :
            wt = math.exp( sample.logW )
            sw += wt
            ws = wt * sample.hyper
            hypar = hypar + ws
            hydev = hydev + ws * sample.hyper
        self.stdevHypars = numpy.sqrt( hydev - hypar * hypar )
        self.hypars = hypar
        return self.hypars

    def getNuisance( self ) :
        """
        Return the average of the nuisance parameters (if present)
        """
        ( a, s ) = self.averstd( "nuisance" )
        self.stdevNuis = s
        return a

    def averstd( self, name ) :
        """
        Return the average and the stddevs of the named attribute from Sample

        Parameters
        ----------
        name : str
            name of an attribute from Sample
        """
        np = len( getattr( self[0], name ) )

        aver = numpy.zeros( np, dtype=float )
        stdv = numpy.zeros( np, dtype=float )
        sw = 0.0
        for sample in self :
            wt = math.exp( sample.logW )
            sw += wt
            ss = getattr( sample, name )
            ws = wt * ss
            aver = aver + ws
            stdv = stdv + ws * ss
        stdv = numpy.sqrt( stdv - aver * aver )

        return ( aver, stdv )

    # ===== MEDIAN ===========================================================
    def getMedianIndex( self ) :
        """
        Return the index at which the median can be found.
        """
        sum = 0.0
        k = 0
        for sample in self :
            sum += math.exp( sample.logW )
            if sum < 0.5 :
                k += 1
            else :
                break
        self.medianIndex = k
        return self.medianIndex

     # ===== EVOLUTIONS ========================================================
    def getMaximumNumberOfParameters( self ):
        """
        Return the maximum number of parameters (for Dynamic Models)
        """
        if self[0].model.isDynamic( ) :
            return numpy.max( self.getNumberOfParametersEvolution() )

        return self[0].model.npchain

    def getParameterEvolution( self, kpar=None ):
        """
        Return the evolution of one or all parameters.

        In case of dynamic models the number of parameters may vary.
        They are zero-padded. Use `getNumberOfParametersEvolution`
        to get the actual number.

        Parameters
        ----------
        kpar : int or tuple of ints
            the parameter to be selected. Default: all

        """
        pe = []
        for sample in self :
            pe += [sample.parameters]
        if kpar is None :
            return numpy.asarray( pe )
        else :
            return numpy.asarray( pe )[:,kpar]

    def getParAndWgtEvolution( self ):
        """
        Return the evolution of parameters and weights.

        In case of dynamic models the number of parameters may vary.
        They are zero-padded. Use `getNumberOfParametersEvolution`
        to get the actual number.

        """
        pe = []
        we = []
        for sample in self :
            pe += [sample.parameters]
            we += [sample.logW]
        return ( numpy.asarray( pe ), numpy.exp( numpy.asarray( we ) ) )


    def getNumberOfParametersEvolution( self ):
        """ Return the evolution of the number of parameters.  """
        pe = [sample.model.npchain for sample in self]
        return numpy.asarray( pe )

    def getScaleEvolution( self ):
        """ Return the evolution of the scale.  """
        pe = [sample.hypars for sample in self]
        return numpy.asarray( pe )

    def getLogLikelihoodEvolution( self ):
        """ Return the evolution of the log( Likelihood ).  """
        pe = [sample.logL for sample in self]
        return numpy.asarray( pe )

    def getLogWeightEvolution( self ):
        """
        Return the evolution of the log( weight ).

        The weights itself sum up to 1.
        See #getWeightEvolution( ).

        """
        pe = [sample.logW for sample in self]
        return numpy.asarray( pe )

    def getWeightEvolution( self ):
        """
        Return the evolution of the weight.

        The weights sum to 1.

        """
        return numpy.exp( self.getLogWeightEvolution( ) )

    def getParentEvolution( self ):
        """ Return the evolution of the parentage.  """
        pe = []
        for sample in self :
            pe += [sample.parent]
        return numpy.asarray( pe )

    def getStartEvolution( self ):
        """ Return the evolution of the start generation.  """
        pe = []
        for sample in self :
            pe += [sample.start]
        return numpy.asarray( pe )

    def getGeneration( self ):
        """ Return the generation number pertaining to the evolution.  """
        pe = []
        for sample in self :
            pe += [sample.id]
        return numpy.asarray( pe )

    def getLowLogL( self ):
        """
        Return the lowest value of logL in the samplelist, plus its index.
        """
        low = self[0].logL
        klo = 0
        k = 0
        for sample in self :
            if sample.logL < low :
                klo = k
                low = sample.logL
            k += 1
        return ( low, klo )


    # ===== AVERAGE RESULTS ===================================================
    def average( self, xdata ):
        """
        Return the (weighted) average result of the model(s) over the samples.

        Parameters
        ----------
        xdata : array_like
            the input

        """
        ndata = Tools.length( xdata )
        result = numpy.zeros( ndata, dtype=float )
        error = numpy.zeros( ndata, dtype=float )
        sumwgt = 0
        for sample in self :
            yfit = sample.model.result( xdata, sample.parameters )
            wgt = math.exp( sample.logW )
            yw = yfit * wgt
            result += yw
            error  += yw * yfit
            sumwgt += wgt

#        self.error = numpy.sqrt( ( error - result * result ) / self.ndata )
        self.error = numpy.sqrt( error - result * result )
        self.result = result

        return self.result

    # ===== MONTE CARLO ERRORS ===================================================
    def monteCarloError( self, xdata ):
        """
        Calculates 1-\sigma-confidence regions on the model given some inputs.

        The model is run with the input for the parameters in each of the
        samples. Appropiately weighted standard deviations are calculated
        and returned at each input value.

        Parameters
        ----------
        xdata : array_like
           the input vectors.

        Returns
        -------
        error : array_like
            standard deviations at each input point

        """
        if Tools.length( xdata ) != len( self.result ):
            self.average( xdata )
        return self.error




import numpy as numpy
import math
from . import Tools

from .MaxLikelihoodFitter import MaxLikelihoodFitter
from .AnnealingAmoeba import AnnealingAmoeba

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2023 Do Kester

class AmoebaFitter( MaxLikelihoodFitter ):
    """
    Fitter using the simulated annealing simplex minimum finding algorithm,

    See also: @AnnealingAmoeba

    Author       Do Kester

    Examples
    --------
    # assume x and y are Double1d data arrays.
    >>> x = numpy.arange( 100, dtype=float ) / 10
    >>> y = 3.5 * SIN( x + 0.4 )                    # make sine
    >>> numpy.random.seed( 12345L )                 # Gaussian random number generator
    >>> y += numpy.random.randn( 100 ) * 0.2        # add noise
    >>> sine = SineModel( )                         # sinusiodal model
    >>> lolim = numpy.asarray( [1,-10,-10], dtype=float )
    >>> hilim = numpy.asarray( [100,10,10], dtype=float )
    >>> sine.setLimits( lolim, hilim )              # set limits on the model parameters
    >>> amfit = AmoebaFitter( x, sine )
    >>> param = amfit.fit( y, temp=10 )
    >>> stdev = amfit.getStandardDeviation( )       # stdevs on the parameters
    >>> chisq = amfit.getChiSquared( )
    >>> scale = amfit.getScale( )                 # noise scale
    >>> yfit  = amfit.getResult( )                # fitted values
    >>> yfit  = sine( x )                         # fitted values ( same as previous )
    >>> yband = amfit.monteCarloError( )               # 1 sigma confidence region
    # for diagnostics ( or just for fun )
    >>> amfit = AmoebaFitter( x, sine )
    >>> amfit.setTemperature( 10 )                # set a temperature to escape local minima
    >>> amfit.setVerbose( 10 )                    # report every 10th iteration
    >>> plotter = IterationPlotter( )             # from BayesicFitting
    >>> amfit.setPlotter( plotter, 20 )            # make a plot every 20th iteration
    >>> param = amfit.fit( y )


    Notes
    -----
    1. AmoebaFitter is not guaranteed to find the global minimum.
    2. The calculation of the evidence is an Gaussian approximation which is
    only exact for linear models with a fixed scale.

    Author : Do Kester.

    """
    #  *************************************************************************
    def __init__( self, xdata, model, **kwargs ):
        """
        Create a new Amoeba class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            independent input values
        model : Model
            the model function to be fitted
        kwargs : dict
            Possibly includes keywords from
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        if model.npchain <= 1 :
            raise ValueError( "AmoebaFitter cannot make a simplex of one parameter" )

        super( AmoebaFitter, self ).__init__( xdata, model, **kwargs )

    #  *************************************************************************
    def fit( self, data, weights=None, par0=None, keep=None, size=None,
                    seed=4567, temp=0, limits=None, maxiter=1000,
                    tolerance=0.0001, cooling=0.95, steps=10,
                    verbose=0, plot=False, accuracy=None, callback=None ):
### TBC parameter defaults


        """
        Return Model fitted to the data array.

        When done, it also calculates the hessian matrix and chisq.

        Parameters
        ----------
        data : array_like
             the data vector to be fitted
        weights : array_like
            weights pertaining to the data
            The weights are relative weights unless `scale` is set.
        accuracy : float or array_like
            accuracy of (individual) data
        par0 : array_like
            initial values of teh parameters of the model
            default: from model
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep are only valid for *this* fit
            See also `AmoebaFitter( ..., keep=dict )`
        size : float or array_like
            step size of the simplex
        seed : int
            for random number generator
        temp : float
            temperature of annealing (0 is no annealing)
        limits : None or list of 2 floats or list of 2 array_like
            None : no limits applied
            [lo,hi] : low and high limits for all values
            [la,ha] : low array and high array limits for the values
        maxiter : int
            max number of iterations
        tolerance : float
            stops when ( |hi-lo| / (|hi|+|lo|) ) < tolerance
        cooling : float
            cooling factor when annealing
        steps : int
            number of cycles in each cooling step.
        verbose : int
            0 : silent
            1 : print results to output
            2 : print some info every 100 iterations
            3 : print some info all iterations
        plot : bool
            plot the results.
        callback : callable
            is called each iteration as
            `val = callback( val )`
            where `val` is the minimizable array

        """
        fitIndex, data, weights = self.fitprolog( data, weights=weights, 
                        accuracy=accuracy, keep=keep )

        func = self.makeFuncs( data, weights=weights, index=fitIndex, ret=1 )

        if par0 is None :
            par0 = self.model.parameters
        if fitIndex is not None and len( fitIndex ) < len( par0 ) :
            par0 = par0[fitIndex]

        kwargs = {}
        if size is not None :
            kwargs["size"] = size
        if seed is not None :
            kwargs["seed"] = seed
        if temp is not None :
            kwargs["temp"] = temp
        if limits is not None :
            kwargs["limits"] = limits
        if maxiter is not None :
            kwargs["maxiter"] = maxiter
        if tolerance is not None :
            kwargs["reltol"] = tolerance
            kwargs["abstol"] = tolerance
        if cooling is not None :
            kwargs["cooling"] = cooling
        if steps is not None :
            kwargs["steps"] = steps
        if verbose is not None :
            kwargs["verbose"] = verbose
        if callback is not None :
            kwargs["callback"] = callback

        amoeba = AnnealingAmoeba( func, par0, **kwargs )

        par = amoeba.minimize()
        parameters = self.insertParameters( par, index=fitIndex )
        self.model.parameters = parameters

        if self.isChisq :
            self.chisq = amoeba.fopt
        else :
            self.logLikelihood = -amoeba.fopt
            self.chisq = self.chiSquared( data, weights=weights )

        self.iter = amoeba.iter

        self.ntrans = amoeba.ncalls
        self.simplex = amoeba.simplex
        self.values = amoeba.values

#        plot = plot or ( verbose == 2 )
        self.fitpostscript( data, plot=plot )

        return parameters

    def __str__( self ):
        """ Return name of the fitter.  """
        return "AmoebaFitter"





import numpy as numpy
import re

from .Tools import setAttribute as setatt
from .Tools import printclass as printclass
from .Dynamic import Dynamic

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2023 Do Kester

class Problem( object ):
    """
    Problem implements the common parts of specialized Problems.

    A Problem is an optimization of parameters which does not involve
    the fitting of data to a Model.

    Problems can be solved by NestedSampler, with appropriate Engines and
    ErrorDistributions.

    The result of the function for certain x and p is given by
    problem.result( x, p )
    The parameters, p, are to be optimized while the x provide additional
    information.

    This class is a base class. Further specializations will define the
    result method.

    Attributes
    ----------
    model : Model
        to be optimized
    xdata : array_like
        independent variable (static)
    ydata : array_like
        dependent variable (static)
    weights : array_like
        weights associated with ydata
    varyy : float or ndarry of shape (ndata,)
        Variance in the errors of the ydata
    npars : int
        number of parameters in the model of the problem
    partype : float | int
        type of the parameters


    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, model=None, xdata=None, ydata=None, weights=None, 
                  accuracy=None, copy=None ):
        """
        Problem Constructor.

        Parameters
        ----------
        model : Model
            the model to be solved
        xdata : array_like or None
            independent variable
        ydata : array_like or None
            dependent variable
        weights : array_like or None
            weights associated with ydata
        accuracy : float or ndarray of shape (ndata,)
            accuracy scale for the datapoints
            all the same or one for each data point
        copy : Problem
            to be copied

        """
        super( ).__init__( )

        if copy is None :
            self.xdata = None if xdata is None else numpy.asarray( xdata )
            self.ydata = None if ydata is None else numpy.asarray( ydata )
            self.model = model
            self.weights = None if weights is None else numpy.asarray( weights )
            self.partype = float
            self.setAccuracy( accuracy=accuracy )
        else :
            self.xdata = copy.xdata
            self.ydata = copy.ydata
            self.weights = copy.weights
            self.varyy = copy.varyy
            self.accuracy = copy.accuracy
            self.hasAccuracy = copy.hasAccuracy
            if copy.model and ( copy.model.isDynamic() or copy.model.isModifiable() ) :
                self.model = copy.model.copy()
            else :
                self.model = copy.model
            self.partype = copy.partype

        if self.model is None or not hasattr( self.model, "cyclic" ) :
            pass                                    # default is no correction
        elif isinstance( self.model.cyclic, dict ) :
            self.cyclicCorrection = self.cycor2
        else :
            self.cyclicCorrection = self.cycor1

    def copy( self ):
        """
        Copy.

        """
        return Problem( copy=self )


    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        setatt( self, name, value )
        if name == "weights" :
            try :
                delattr( self, "sumweight" )
            except :
                pass
        else :
            setatt( self, name, value )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == 'npars' :
            return self.model.npars
        elif name == 'varyy' :
            return 0
        elif name == 'sumweight' :            # Return the sum over weight vector.
            if self.hasWeights() :
                self.sumweight = numpy.sum( self.weights )
                return self.sumweight
            else :
                return self.ndata
        elif name == 'ndata' :
            self.ndata = len( self.ydata )        # number of data points/tuples
            return self.ndata
        elif name == 'logScale' :
            return 0.0
        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    def setAccuracy( self, accuracy=None ) :
        """
        set the value for accuracy.

        Paramaters
        ----------
        accuracy : float or array of NDATA floats or None
            Either one value for all or one for each data point
            When None the value is set to 0 (for easy computational reasons)
        """
        self.hasAccuracy = accuracy is not None
        self.accuracy = accuracy if self.hasAccuracy else 0
        self.varyy = self.accuracy * self.accuracy

    def hasWeights( self ):
        """ Return whether it has weights.  """
        return self.weights is not None


    #  *****RESULT**************************************************************
    def result( self, param ):
        """
        Returns the result using the parameters.

        In this (base)class it is a placeholder.

        Parameters
        ----------
        param : array_like
            values for the parameters.

        """
        pass


    def residuals( self, param, mockdata=None ) :
        """
        Returns the residuals, calculated at the xdata.

        Parameters
        ----------
        param : array_like
            values for the parameters.
        mockdata : array_like
            model fit at xdata

        """
        res = self.ydata - ( self.result( param ) if mockdata is None else mockdata )
        return self.cyclicCorrection( res )


    def cyclicCorrection( self, res ):
        """
        No correction.

        Returns 
        -------
            the residuals, unadultered
        
        Parameters
        ----------
        res : array_like
            residuals

        """
        return res

    def cycor1( self, res ):
        """
        Returns the residuals, all corrected for periodicity in residuals

        Parameters
        ----------
        res : array_like
            residuals

        """
        return self.cyclize( res, self.model.cyclic )

    def cycor2( self, res ):
        """
        Returns the residuals corrected for periodicity in residuals, only
        the result dimensions listed in the model.cyclic dictionary.

        Parameters
        ----------
        res : array_like
            residuals

        """
        cyclic = self.model.cyclic
        for key in cyclic.keys() :
            res[:,key] = self.cyclize( res[:,key], cyclic[key] )
        return res

    def cyclize( self, res, period ) :
        """
        Apply correction on residuals which are cyclic in some
        phase space.

        If the model results in a phase value of +&epsilon;
        while the data give that phase value as (p - &epsilon;)
        to keep all data in the range [0,p], the naive residual
        would be (p - 2 &epsilon;) while the actual distance should
        be measured the other way around as (2 &epsilon;).
        Here p = period and &epsilon; = small deviation.

        Parameters
        ----------
        res : array_like
            original residuals
        period : float
            of the phase space

        Returns
        -------
        corrected residuals.
        """

        hp = period / 2
        return numpy.where( res < -hp, res + period,
               numpy.where( res >  hp, res - period, res ) )



    def XXXweightedResiduals( self, param, mockdata=None, extra=False ) :
        """
        Returns the (weighted) residuals, calculated at the xdata.

        Optionally (extra=True) the weighted signs of the residuals are returned too.

        Parameters
        ----------
        param : array_like
            values for the parameters.
        mockdata : array_like
            model fit at xdata
        extra : bool (False)
            true  : return ( wgt * res, wgt * sign( res ) )
            false : return wgt * res
        """
        res = self.residuals( param, mockdata=mockdata )

        if extra :
            if self.weights is None :
                wgt = numpy.ones_like( self.ydata )
                return ( res, numpy.copysign( wgt, res ) )
            else :
                return ( res * self.weights, numpy.copysign( self.weights, res ) )
        else :
            return res if self.weights is None else res * self.weights


    def weightedResSq( self, allpars, mockdata=None, extra=False ) :
        """
        Returns the (weighted) squared residuals, calculated at the xdata.

        Optionally (extra=True) the weighted residuals themselves are returned too.

        Parameters
        ----------
        allpars : array_like
            values for the parameters.
        mockdata : array_like
            model fit at xdata
        extra : bool (False)
            true  : return ( wgt * res^2, wgt * res )
            false : return wgt * res^2
        """
        res = self.residuals( allpars[:self.npars], mockdata=mockdata )

        resw = res if self.weights is None else res * self.weights
        return ( resw * res, resw ) if extra else resw * res

    def isDynamic( self ) :
        return self.model.isDynamic()


    def domain2Unit( self, dval, kpar ) :
        """
        Return value in [0,1] for the selected parameter.

        Parameters
        ----------
        dval : float
            domain value for the selected parameter
        kpar : array_like
            selected parameter index, where kp is index in [parameters, hyperparams]
        """
        return self.model.domain2Unit( dval, kpar )


    def unit2Domain( self, uval, kpar ) :
        """
        Return domain value for the selected parameter.

        Parameters
        ----------
        uval : array_like
            unit value for the selected parameter
        kpar : array_like
            selected parameter indices, where kp is index in [parameters, hyperparams]
        """
        return self.model.unit2Domain( uval, kpar )

    #  *****TOSTRING***********************************************************
    def __str__( self ):
        """ Returns a string representation of the model.  """
        return self.baseName( )

    def _toString( self, spaces ) :
        return self.baseName()


    def shortName( self ):
        """
        Return a short version the string representation: upto first non-letter.

        """
        m = re.match( "^[a-zA-Z_]*", self.baseName() )
        return m.group(0)

    def baseName( self ) :
        return "Problem of %s" % self.model.shortName()




import numpy as numpy
import numpy.linalg
from astropy import units
import math

from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt
from .LinearModel import LinearModel
from .kernels.Kernel import Kernel
from .kernels.Tophat import Tophat

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2022 Do Kester

class FreeShapeModel( LinearModel ):
    """
    Pixelated Model.

    f( x:p ) = p( expanded )

    where p is a array of amplitudes of the same size as the input x divided by
    the number of pixels per bin ( ppb ). When ppb > 1, each p has a shape which
    is used to fill the pixels of the bin. Initially the shape is a top-hat,
    which can be autoconvolved.

    By default ppb = 5.

    The parameters are initialized at {0}.

    Although this is a LinearModel it will not work very well with the ( linear )
    Fitter. It will be a very ill-posed problem.

    Using NestedSampler its exponential prior will ensure that all
    parameters are kept positive.

    Attributes
    ----------
    npix : int
        Number of pixels in result. Is also npar.
    xlo : float ( default 0 )
        Lowest value in xdata
    xhi : float ( default npix )
        Highest value in xdata
        xlo and xhi define the valid domain of the model.
        All input data must be: xlo <= xdata <= xhi
    shape : Kernel
        shape of convolving function
    center : float (between 0..1)
        position of the center of shape with respect to the pixels

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Examples
    --------
    >>> nn = 100
    >>> fsm = FreeShapeModel( nn, nconvolve=4, xlo=-1.0, xhi=4.0 )
    >>> print( fsm.shape )

    Author       Do Kester

    """
    def __init__( self, npix, copy=None, shape=None, nconvolve=0,
                  center=0.5, xlo=0.0, xhi=None, **kwargs ):
        """
        Free Shape model with npix pixels.

        The number of parameters equals the number of pixels

        Parameters
        ----------
        npix : int
            number of pixels = npar
        copy : FreeShapeModel
            model to be copied
        shape : None or Kernel
            None : Use Tophat(), convolved nconvolve times.
            Kernel : use the kernel as shape; nconvolve does not apply.
        nconvolve : int
            number of (auto)convolutions on Tophat
        center : float (between 0..1)
            positions where the pixels are centered.
            default: 0.5 -> pixels run from k to k+1
        xlo : float ( default 0.0 )
            lowest value in xdata
        xhi : float ( default np )
            highest value in xdata

        """
        super( FreeShapeModel, self ).__init__( npix, copy=copy )

        if copy is None :
            if shape is None :
                setatt( self, "shape", Tophat( nconv=nconvolve ) )
            else :
                self.shape = shape
            self.xlo = xlo
            self.xhi = xhi
            self.center = center

        else :
            setatt( self, "xlo", copy.xlo )
            setatt( self, "xhi", copy.xhi )
            setatt( self, "shape", copy.shape )
            setatt( self, "center", copy.center )

    def copy( self ):
        """ Copy method.  """
        return FreeShapeModel( self.npbase,  copy=self )

    def __setattr__( self, name, value ) :
        if name == "shape" :
            setatt( self, name, value, type=Kernel )
            return
        if name == "xlo" :
            setatt( self, name, value, type=float )
            return
        if name == "xhi" :
            setatt( self, name, self.npbase if value is None else value, type=float )
            return
        if name == "center" and ( 0 <= value <= 1 ) :
            setatt( self, name, value, type=float )
            return

        super( FreeShapeModel, self ).__setattr__( name, value )


    def checkDomain( self, xdata ) :
        """
        Check for all data inside domain defined by (xlo - range, xhi + range).
        range = self.shape.range

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result

        Raises
        ------
        ValueError when outside domain.
        """
        if numpy.any( numpy.logical_or( xdata < self.xlo - self.shape.range,
                                        xdata > self.xhi + self.shape.range ) ) :
            raise ValueError( "Some values are outside domain (%f,%f)" %
                                (self.xlo, self.xhi) )


    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        self.checkDomain( xdata )

        return super( FreeShapeModel, self ).baseResult( xdata, params )

    def basePartial( self, xdata, params, parlist=None ) :
        """
        Returns the partial derivative of the model function to
        each of the parameters.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
             values for the parameters

        """
        self.checkDomain( xdata )

        nxdata = Tools.length( xdata )
        xpix = self.npbase * ( xdata - self.xlo ) / ( self.xhi - self.xlo )
        xpix += self.center

        if parlist is None :
            parlist = [k for k in range( self.npbase )]
        np = len( parlist )

        partial = numpy.zeros( ( nxdata, np ), dtype=float )
        p0 = 0
        for k,p in enumerate( parlist ) :
            xpix -= ( p - p0 )
            p0 = p
            partial[:,k] = self.shape.result( xpix )

#        print( fmt( partial ) )
        return partial

    def TBCbaseDerivative( self, xdata, params ):
        """
        Returns the derivative of the model function df/dx.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
             values for the parameters

        """
        nxdata = Tools.length( xdata )
        xpix = self.npbase * ( xdata - self.xlo ) / ( self.xhi - self.xlo )
        xpix += self.center

        partial = numpy.zeros( ( nxdata, self.npbase ), dtype=float )
        for k in range( self.npbase ) :
            partial[:,k] = self.shape.partial( xpix )
            xpix -= 1

        return numpy.inner( params, partial )

    def baseName( self ):
        return str( "FreeShape with %d pixels." % self.npbase )

    def baseParameterUnit(  self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        return self.yUnit



import numpy as numpy
from .LinearModel import LinearModel
from . import Tools
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.8.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA Code)
#  *    2016 - 2021 Do Kester


class PolynomialModel( LinearModel ):
    """
    General polynomial model of arbitrary degree.

        f( x:p ) = &sum; p_k * x^k

    where the sum is over k running from 0 to degree ( inclusive ).

    It is a linear model.
    Examples
    --------
    >>> poly = PolynomialModel( 3 )            # 3rd degree polynomial
    >>> print( poly.getNumberOfParameters() )
    4

    Author : Do Kester

    Attributes
    ----------
    degree : int
        degree of the polynomial

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """
    def __init__( self, degree, copy=None, **kwargs ):
        """
        Polynomial of a certain degree.

        The number of parameters is ( degree + 1 )

        Parameters
        ----------
        degree : int
            the degree of the polynomial.
        copy : PolynomialModel
            model to copy
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        super(PolynomialModel,self ).__init__( degree + 1, copy=copy,
                        **kwargs )

        self.degree = degree

    def copy( self ):
        """ Copy method.  """
        return PolynomialModel( self.degree, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: degree
        """
        if name == 'degree' :
            setatt( self, name, value, type=int )
        else :
            super( PolynomialModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        The partials are the powers of x ( xdata ) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model (ignored for LinearModels).
        parlist : array_like
            list of indices of active parameters

        """
        nxdata = Tools.length( xdata )
        x = numpy.ones( nxdata )

        partial = numpy.zeros( ( nxdata, self.npmax ), dtype=float )
        for k in range( self.npmax ) :
            partial[:,k] = x
            x *= xdata

        if parlist is None :
            return partial

        return partial[:,parlist]

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each input (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model.

        """
        if self.npmax <= 1 :
            return numpy.zeros_like( xdata )

        np = self.npmax
        p = params * numpy.arange( np )
        np -= 2
        return PolynomialModel( np ).result( xdata, p[1:] )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        bn = "Polynomial: f( x:p ) = p_0"
        if self.npmax > 1 :
            bn += " + p_1 * x"
        for k in range( 2, self.npmax ) :
            bn += " + p_%d * x^%d"%(k,k)
        return bn

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.
        Parameters
        ----------
        k : int
            parameter number.

        """
        return "polycoeff_%d" % k

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        """
        return self.yUnit / ( self.xUnit ** k )





import numpy as numpy
from astropy import units
from astropy.modeling import Model as AstroModel
from astropy.modeling import FittableModel
import math
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt

from .Model import Model

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2021        Do Kester


class AstropyModel( Model ):
    """
    Wrapper for Astropy Models, so they can be used in BayesicFitting.

    Examples
    --------
    >>> from astropy.modelling.models import Gaussian1D
    >>> gm = Gaussian1D( mean=0, stddev=1, amplitude=1 )
    >>>
    >>> gauss = AstropyModel( gm )
    >>> print( gauss )
    AstropyModel( Gauss1D )
    >>> print( gauss.getNumberOfParameters( ) )
    3
    >>> print( gauss( numpy.linspace( -5, 5, 11 ) ) )
    [  3.72665317e-06   3.35462628e-04   1.11089965e-02   1.35335283e-01
       6.06530660e-01   1.00000000e+00   6.06530660e-01   1.35335283e-01
       1.11089965e-02   3.35462628e-04   3.72665317e-06]

    Attributes
    ----------
    astromodel : astropy.modeling.FittableModel
        The astropy model to be wrapped

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Alternate
    ---------
    `GaussModel()` is equivalent to `KernelModel( kernel=Gauss() )`.


    """

    def __init__( self, astromodel, copy=None, **kwargs ):
        """
        Gaussian model.

        Number of parameters is 3.

        Parameters
        ----------
        fitmodel : FittableModel
            FittableModel from astropy.modeling  
        copy : GaussModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """

        setatt( self, "astromodel", astromodel, type=FittableModel )
        
        npars = len( astromodel.parameters )
        ndim = astromodel.n_inputs
        ndout = astromodel.n_outputs
        if ndout > 1 :
            setatt( self, "ndout", ndout )

        super( ).__init__( npars, ndim=ndim, names=astromodel.param_names, 
                           params=astromodel.parameters, copy=copy, **kwargs )


    def copy( self ):
        """ Copy method.  """
        return AstropyModel( self.astromodel.copy(), copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        if self.ndim == 1 :
            return self.astromodel.evaluate( xdata, *tuple( params ) )
        else :
            xd = list( xdata.T )
            return self.astromodel.evaluate( *tuple( xd ), *tuple( params ) )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        if self.ndim == 1 :
            ap = self.astromodel.fit_deriv( xdata, *tuple( params ) )
        else :
            xd = list( xdata.T )
            ap = self.astromodel.fit_deriv( *tuple( xd ), *tuple( params ) )
    
        if isinstance( ap, list ) :
            ap = numpy.asarray( ap ).T

        if parlist is None :
            return ap

        partial = numpy.ndarray( ( Tools.length( xdata ), self.npbase ) )
        for k,kp in enumerate( parlist ) :
            partial[:,k] = ap[:,kp]

        return partial

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        self.numDerivative( xdata, params )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        name = str( self.astromodel )
        ke = name.find( "\n" )
        return str( "AstropyModel( %s )" % name[7:ke] )






import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
from BayesicFitting import formatter as fmt

from BayesicFitting import *


class TestPlot( unittest.TestCase ):
    "A few simple tests for HelpPlot"

    def __init__( self, testname ):
        super( ).__init__( testname )
        print( os.environ )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

        print( "DP  ", self.doplot )

    def test( self ):
        print( "==== test 1 ========" )
        a = numpy.arange( 11, dtype=float )
        b = numpy.arange( 11, dtype=float )
#        self.assertEqual(a, b)
        self.assertTrue( True )
        print( "DPi ", self.doplot )
        if self.doplot :
            plt.plot( a, b )
            plt.show()
        else :
            print( "no plot" )

if __name__ == '__main__':
    unittest.main()




import numpy as numpy
from astropy import units
import math
from .import Tools
from .Tools import setAttribute as setatt

from .HarmonicModel import HarmonicModel
from .Dynamic import Dynamic

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2018 - 2020 Do Kester


class HarmonicDynamicModel( HarmonicModel, Dynamic ):
    """
    Harmonic oscillator Model of adaptable order.

    f( x:p ) = &sum;_j ( p_k cos( 2 &pi; j x ) + p_k+1 sin( 2 &pi; j x ) )

    j = 1, N; k = 0, 2N

    The parameters are initialized at 1.0. It is a linear model.

    Author       Do Kester

    Attributes
    ----------
    minOrder : int
        minimum degree of polynomial (def=1)
        Can also be read as minComp
    maxOrder : None or int
        maximum degree of polynomial (def=None)
        Can also be read as maxComp

    Attributes from Dynamic
    ----------------------
        ncomp (= order), deltaNpar, minComp (= minOrder), maxComp (= maxComp), growPrior

    Attributes from HarmonicModel
    -----------------------------
        order, period

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    Examples
    --------
    >>> harm = HarmonicDynamicModel( 3 )            # period = 1
    >>> print harm.getNumberOfParameters( )         # 6
    >>> harm = HarmonicModel( 4, period=2.7 )       # period = 2.7

    Category     mathematics/Fitting

    """
    deltaNpar = 2

    def __init__( self, order, minOrder=1, maxOrder=None, period=1.0, fixed=None,
                  growPrior=None, copy=None, **kwargs ):
        """
        Harmonic of a adaptable order.

        The model starts as a HarmonicModel of order = 1
        Growth of the model is governed by a prior.

        Parameters
        ----------
        order : int
            order to start with. It should be minOrder <= order <= maxOrder
        minOrder : int
            minimum degree of polynomial (def=1)
        maxOrder : None or int
            maximum degree of polynomial (def=None)
        period : float
            period of the oscilation
        fixed : None
            If fixed is not None an AttributeError is raised
        growPrior : None or Prior
            governing the birth and death.
            ExponentialPrior (scale=2) if  maxOrder is None else UniformPrior
        copy : HarmonicDynamicModel
            model to copy

        Raises
        ------
        AttributeError when fixed parameters are requested
        ValueError when order is outside [min..max] range

        """
        if fixed is not None :
            raise AttributeError( "DynamicModel cannot have fixed parameters" )
        if order < minOrder or ( maxOrder is not None and order > maxOrder ) :
            raise ValueError( "order outside range of [min..max] range" )

        super( HarmonicDynamicModel, self ).__init__( order, period=period,
            copy=copy, **kwargs )

        self.deltaNpar = 2

        if copy is None :
            self.setGrowPrior( growPrior=growPrior, min=minOrder, max=maxOrder,
                           name="Order" )
        else :
            setatt( self, "minOrder", copy.minOrder )
            setatt( self, "maxOrder", copy.maxOrder )
            setatt( self, "growPrior", copy.growPrior.copy() )

    def copy( self ):
        """ Copy method.  """
        return HarmonicDynamicModel( self.order, period=self.period, copy=self )

    def isDynamic( self ) :
        return True

    def changeNComp( self, dn ) :
        setatt( self, "order", self.order + dn )

    def __setattr__( self, name, value ) :
        if self.setDynamicAttribute( name, value ) :
            return
        else :
            super( ).__setattr__( name, value )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == 'ncomp' :
            return self.order
        if name == 'minComp' :
            return self.minOrder
        if name == 'maxComp' :
            return self.maxOrder

        return super( ).__getattr__( name )

    def basePrior( self, k ):
        """
        Return the prior for parameter k.

        Parameters
        ----------
        k : int
            the parameter to be selected.
        """
        return super( HarmonicDynamicModel, self ).basePrior( k % self.deltaNpar )


    def baseName( self ):
        """ Return a string representation of the model.  """
        return "Dynamic" + super( HarmonicDynamicModel, self ).baseName( )





import numpy as numpy
from astropy import units
import math
import re
from . import Tools
from .Tools import setAttribute as setatt

from .BracketModel import BracketModel
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2011 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester

class CombiModel( BracketModel ):
    """
    CombiModel combines a number of copies of the same model.

    Fixed relations can be set between similar parameters.
    The relations can be either multiplicative or additive.
    When these relations are set, they must be set for all models.

        f( x:p ) = &sum; g( x:p )

    where g( x:p ) is a model ( e.g. GaussModel )

    For consistency reasons it is not possible to change the attributes of a
    CombiModel. It is better to make a new one with the required settings.

    Attributes
    ----------
    nrepeat : int
        number of Models in this CombiModel
    nmp : int
        number of parameters in each Model
    expandpar : array of float
        expanded parameters. to be used in the repeated models
    npcmax : int
        number of parameters in expandpar (= nmp * nrepeat)
    expandindex : array of int
        converts parameter index into expandpar index
    addindex : array of int
        indices of additively connected parameters
    addvalue : array of float
        list of values to be added to the parameters in addindex
    mulindex : array of int
        indices of multiplicatively connected parameters
    mulvalue : array of float
        list of values to be multiplied to the parameters in mulindex
    select : array of int
        indices of expandpar to get parameters

    Attributes from BracketModel
    --------------------------
        model, deep

    Attributes from Model
    --------------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    Examples
    --------
    >>> gauss = GaussModel( )
    >>> combi = CombiModel( gauss, 3, addCombi={1:[0,0.1,0.3]}, mulCombi={2,[0]*3} )
    >>> print combi
    Combi of 3 times GaussModel
    >>> print( combi.npchain, combi.nrepeat, combi.nmp, combi.nexpand )
    5 3 3 9
    >>> print( combi.select )
    [0 1 2 3 6]
    >>> print( combi.expandindex )
    [0 1 2 3 1 2 4 1 2]
    >>> print( combi.modelindex )
    [0 0 0 1 2]
    >>> print( combi.addindex )
    [1 4 7]
    >>> print( combi.mulindex )
    [2 5 8]

    Category     mathematics/Fitting

    Notes
    -----
    1. When all parameters are left free, precise initial parameters are
    needed to converge to the global optimum.
    2. The model seems to be especially unstable when the basic models
    are overlapping. Fixing the widths relative to each other, helps.
    3. Using a PolynomialModel ( or similar ones ) as basic model, is not going
    to work.

    """
    def __init__( self, model, nrepeat=1, copy=None, oper='add',
                  addCombi=None, mulCombi=None, **kwargs ):
        """
        CombiModel combines several copies of the same model int one.

        Parameters
        ----------
        model : Model
            model to be repeated
        nrepeat : int
            number of repetitions
        oper : "add" or "mul"
            the repeated models are combined using this operation
        addCombi : None or dict
            make additive connections between parameters
            None : no additive connection
            dict : { int : array }
                key : int
                    index pointing to the key-th parameter in the model
                value : array of nrepeat floats
                    values added to each of the key-th parameters
        mulCombi : None or dict
            make multiplicative connections between parameters
            None : no multiplicative connection
            dict : { int : array }
                key : int
                    index pointing to the key-th parameter in the model
                value : array of nrepeat floats
                    values  multiplied to each of the key-th parameters

        """
        mdl = model.copy()
        for k in range( 1, nrepeat ) :
            if oper == 'add' :
                mdl.addModel( model.copy() )
            elif oper == 'mul' :
                mdl.multiplyModel( model.copy() )
            else :
                raise ValueError( "Unknown operation : " + oper )

        super( CombiModel, self ).__init__( mdl, **kwargs )

        setatt( self, "nmp", model.npchain )
        setatt( self, "nrepeat", nrepeat )
        setatt( self, "npcmax", model.npchain * nrepeat )

        self.combine( addCombi=addCombi, mulCombi=mulCombi )

        if copy is not None :
            setatt( self, "addindex", copy.addindex.copy() )
            setatt( self, "addvalue", copy.addvalue.copy() )
            setatt( self, "mulindex", copy.mulindex.copy() )
            setatt( self, "mulvalue", copy.mulvalue.copy() )
            setatt( self, "select", copy.select.copy() )
            setatt( self, "expandindex", copy.expandindex.copy() )

        setatt( self, "_npchain", len( self.select ) )
        setatt( self, "npbase", self._npchain )
        setatt( self, "npmax", self.npbase )
        setatt( self, "parameters", self.parameters[self.select] )

    def copy( self ):
        """ Copy method.  """
        oper = "add"
        if self.model._next and self.model._next._operation == 3 :
            oper = "mul"

        return CombiModel( self.model.isolateModel( 0 ), nrepeat=self.nrepeat,
                    oper=oper, copy=self )

    def combine( self, addCombi=None, mulCombi=None ) :
        """
        (re)sets the value of attributes "addindex", "addvalue", "mulindex", "mulvalue",
        "select" and "expandindex".

        Parameters
        ----------
        addCombi : None or dict
            make additive connections between parameters
            None : no additive connection
            dict : { int : array }
                key : int
                    index pointing to the key-th parameter in the model
                value : array of nrepeat floats
                    values added to each of the key-th parameters
        mulCombi : None or dict
            make multiplicative connections between parameters
            None : no multiplicative connection
            dict : { int : array }
                key : int
                    index pointing to the key-th parameter in the model
                value : array of nrepeat floats
                    values  multiplied to each of the key-th parameters

        """
#        print( addCombi, mulCombi )

        addindex, addvalue = self.setCombi( addCombi )
        setatt( self, "addindex", addindex )
        setatt( self, "addvalue", addvalue )
        mulindex, mulvalue = self.setCombi( mulCombi )
        setatt( self, "mulindex", mulindex )
        setatt( self, "mulvalue", mulvalue )

        expandindex = numpy.arange( self.npcmax, dtype=int )
        expandindex = self.makeExpandIndex( expandindex, self.addindex )
        expandindex = self.makeExpandIndex( expandindex, self.mulindex )

        setatt( self, "select", numpy.asarray( list( sorted( set( expandindex ) ) ) ) )

        i = 0
        for k in range( self.npcmax ) :
            if expandindex[k] >= i :
                expandindex[k] = i
                i += 1

        setatt( self, "expandindex", expandindex )


    def makeExpandIndex( self, expandindex, amindex ) :
        """
        Make an expanded index enumerating the parameters for the full model
        """

        for i,k in enumerate( amindex ) :
            if i % self.nrepeat == 0 :
                kk = amindex[i]
            else :
                expandindex[k] = kk
#            print( i, k, kk, amindex, expandindex )
        return expandindex


    def setCombi( self, combi ) :
        index = numpy.ndarray( 0, dtype=int )
        value = numpy.ndarray( 0, dtype=float )
        if combi is None :
            return (index,value)
        if not isinstance( combi, dict ) :
            raise ValueError( "The parameters addCombi, mulCombi need to be a dictionary" )

        for key in combi.keys() :
            if key >= self.nmp :
                raise ValueError( "The keys of a combi-dict need to be " +
                                    "< %d"%self.nmp )
            if len( combi[key] ) != self.nrepeat :
                raise ValueError( "The values of a combi-dict need to be " +
                                    "a list of length %d"%self.nrepeat )
            index = numpy.append( index, [k for k in range( key, self.npcmax, self.nmp )] )
            value = numpy.append( value, list( combi[key] ) )
        return ( index, value )

    def expandParameters( self, param ) :
        exppar = param[self.expandindex]
        exppar[self.addindex] += self.addvalue
        exppar[self.mulindex] *= self.mulvalue
        return exppar

    def baseResult( self, xdata, params ) :
        """
        Returns the result calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        exppar = self.expandParameters( params )
        return super( CombiModel, self ).baseResult( xdata, exppar )

    def basePartial( self, xdata, params, parlist=None ) :
        """
        Returns the partial derivatives calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            Not in use

        """
        exppar = self.expandParameters( params )
        fulldp = super( CombiModel, self ).basePartial( xdata, exppar )
        partial = fulldp[:,self.select]
        kk = 0
        for i,k in enumerate( self.addindex ) :
            if i % self.nrepeat == 0 :
                kk = k
            else :
                partial[:,kk] += fulldp[:,k]
        for i,k in enumerate( self.mulindex ) :
            if i % self.nrepeat == 0 :
                kk = k
                partial[:,kk] *= self.mulvalue[i]
            else :
                partial[:,kk] += fulldp[:,k] * self.mulvalue[i]
        return partial

    def baseDerivative( self, xdata, params ) :
        """
        Returns the derivative (df/dx) calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.

        """
        exppar = self.expandParameters( params )
        return super( CombiModel, self ).baseDerivative( xdata, exppar )

    def baseName( self ):
        """ Returns a string representation of the model.  """
        shortname = self.model.shortName( )
        m = re.match( "^[a-zA-Z_]*", shortname )
        return str( "Combi of %d times "%self.nrepeat + m.group(0) )

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        return ( self.model.getParameterName( self.expandindex[k] % self.nmp ) +
                "_%d" % ( self.expandindex[k] / self.nmp ) )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        return self.model.getParameterUnit( self.expandindex[k] )






import numpy as numpy
from astropy import units
import re
import warnings
from . import Tools
from .Tools import setAttribute as setatt

from .Problem import Problem
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018        Do Kester

class ErrorsInXandYProblem( Problem ):
    """
    A ErrorsInXandYProblem is an optimization of parameters which involves
    the fitting of data to a Model, where both the ydata and the xdata
    contain errors.

    It entails that the xdata are not at the exact locations. They need to
    be optimized too. Consequently the parameters of the model are appended
    with a set of parameters, the length of xdata. These extra parameters
    will contain the target locations of the xdata. The 2-dimensional distance
    between data (xdata,ydata) and (target, model(target)) is minimised.
    The target are nuisance parameters which are not part of the modeling
    solution.

    Define
        xd = xdata, yd = ydata, u = target, F(u:P) = model( target )
    And the mismathes in both directions.
        X = u - xd 
        Y = F(u:p) - yd

    Both distances need to be minimized, possibly in the presence of a correlation 
    between the mismatches X and Y 

    As the targets need to be optimised they need a Prior. In the present
    implementation there is the same Prior for all targets, which the centered on
    each of the xdata values.
    S.Gull (1989) argues to use a GaussPrior with a scale similar to the errors
    in both X and Y.

    Attributes
    ----------
    prior : Prior
        Priors for the x-axis nuisance parameters.
    varxx : float or ndarray of shape (ndata,)
        Variance in the xdata errors
    varyy : float or ndarray of shape (ndata,)
        Variance in the ydata errors
    varxy : float or ndarray of shape (ndata,)
        Covariance in the xdata and ydata errors



    Attributes from Problem
    -----------------------
    model, xdata, ydata, weights, partype


    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, model=None, xdata=None, ydata=None, weights=None,
                    prior=None, covar=None, accuracy=None, copy=None ):
        """
        Problem Constructor.

        Parameters
        ----------
        model : Model
            the model to be solved
        xdata : array_like or None
            independent variable
        ydata : array_like or None
            dependent variable
        weights : array_like or None
            weights associated with data
        covar : ndarray of shape (2,2) or (ndata,2,2)
            covariance matrix of the errors in x and y
            (2,2) : valid for all datapoints
            (ndata,2,2): one for each datpoint
            Default is [[1,0],[0,1]]
        accuracy : ndarray of shape (2,) or (3,) or (ndata,2) or (ndata,3)
            accuracy scale for the datapoints
            (2,) scale for resp. y and x, valid for all datapoints
            (3,) scale for y and y, and correlation coefficient between y and x, valid for all
            (ndata,2) or (ndata,3) one set of values for each datapoint
            Alternative for covarince matrix. 
            covar = [[ acc[0]^2, 0], [0, acc[1]^2]] no correlation
                  = [[ acc[0]^2, r], [r, acc[1]^2]] where r = acc[0] * acc[1] * acc[2]
            accuracy is converted to covar; default is covar.
        prior : Prior
            prior for the x-axis nuisance parameters. All centered on each of the xdata
        copy : Problem
            to be copied

        """
        super( ).__init__( model=model, xdata=xdata, ydata=ydata, weights=weights, copy=copy )
        self.npars += Tools.length( self.xdata )

        if copy is None :
            self.prior = prior
            self.setAccuracy( accuracy=accuracy, covar=covar )
        else :
            self.prior = copy.prior
            self.varxx = copy.varxx
            self.varxy = copy.varxy
            self.varyy = copy.varyy
            
            self.determinant = copy.determinant
            self.sumweight = copy.sumweight

    def copy( self ):
        """
        Copy.

        """
        return ErrorsInXandYProblem( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        if name == "accuracy" :
            self.setAccuracy( accuracy=value )
        elif name == "covar" :
            self.setAccuracy( covar=value )
        else :
            super().__setattr__( name, value )


    def setAccuracy( self, accuracy=None, covar=None ) :
        """
        Store 3 items from the covar matrix : 

            | var_yy, var_xy |
            | var_xy, var_xx |

        When the accuracy is given, convert it to these items by
        var_yy = acc[0] * acc[0]
        var_xx = acc[1] * acc[1]
        var_xy = acc[0] * acc[1] * acc[2]

        Store also the determinant of the covariance matrix. 

        When both accuracy and covar are None
          var_yy = 0 
          var_xx = 1
          var_xy = 0

        Raises
        ------
        AttributeError. When both accuracy and covar are not None.

        Parameters
        ----------
        accuracy : ndarray of shape (2,) or (3,) or (ndata,2) or (ndata,3)
            accuracy scale for the datapoints
            (2,) scale for resp. y and x, valid for all datapoints
            (3,) scale for y and y, and correlation coefficient between y and x, valid for all
            (ndata,2) or (ndata,3) one set of values for each datapoint
            Alternative for covarince matrix. 
            vxy = 0 if no correlation else acc[0] * acc[1] * acc[2]
            covar = [[ acc[0]^2, vxy      ],
                     [ vxy     , acc[1]^2 ]] 
            accuracy is converted to covar; default is covar.
        covar : ndarray of shape (2,2) or (ndata,2,2) or None
            covariance matrix of the errors in x and y

        """
        hasAcc = True
        if accuracy is None :
            if  covar is None :
                varyy = 0.0         # scale will fill the value in this case
                varxx = 0.0
                varxy = 0.0
                hasAcc = False
            else :
                covar = numpy.asarray( covar )
                cdim = covar.ndim
                varyy = covar[0,0] if cdim == 2 else covar[:,0,0]
                varxx = covar[1,1] if cdim == 2 else covar[:,1,1]
                varxy = covar[0,1] if cdim == 2 else covar[:,0,1]
        else :
            if covar is None :
                accuracy = numpy.asarray( accuracy )
                adim = accuracy.ndim
#                print( "SetAcc  ", accuracy, adim )
                varxy = 0.0
                if adim == 0 :
                    varyy = accuracy
                    varxx = accuracy
                elif adim == 1 :
                    varyy = accuracy[0] 
                    varxx = accuracy[1]
                    if accuracy.shape[-1] == 3 :
                        varxy = accuracy[2]
                else :
                    varyy = accuracy[:,0]
                    varxx = accuracy[:,1]
                    if accuracy.shape[-1] == 3 :
                        varxy = accuracy[:,2]

                varxy *= varyy * varxx
                varyy *= varyy
                varxx *= varxx
            else :
                raise AttributeError( "Covar and Accuracy are mutually exclusive" )

        setatt( self, "hasAccuracy", hasAcc )
        setatt( self, "accuracy", accuracy if self.hasAccuracy else 0 )
        setatt( self, "varyy", varyy )
        setatt( self, "varxx", varxx )
        setatt( self, "varxy", varxy )

        det = varyy * varxx - varxy * varxy
        setatt( self, "determinant", det )


    def hasWeights( self ):
        """ Return whether it has weights.  """
        return self.weights is not None


    #  *****RESULT**************************************************************
    def result( self, param ):
        """
        Returns the result calculated at the xdatas.

        Parameters
        ----------
        param : array_like
            values for the parameters + nuisance params.

        """
        ( xd, pars ) = self.splitParam( param )
        return self.model.result( xd, pars )


    def splitParam( self, param ) :
        """
        Split the parameters into Model parameters and targets.

        Parameters
        ----------
        param : array_like
            values for the parameters + nuisance params.

        Return
        ------
        tuple of ( targets, model parameters )

        """
        np = self.model.npars
        return ( param[np:], param[:np] )


    def partial( self, param ) :
        """
        Return the partials as a matrix [2*nx,np+nx], where nx is the number of
        datapoints and np the number of parameters in the model.

            The upper left submatrix (nx,np) contains dM/dp
            the upper right submatrix (nx,nx) contains dM/dx on the diagonal
            the lower left submatrix (nx,np) contains zeros
            the lower right submatrix (nx,nx) contains the identity matrix

        Parameters
        ----------
        param : array_like
            values for the parameters + nuisance params.

        """
        ( xd, pars ) = self.splitParam( param )
        np = self.model.npars
        nx = len( xd )

        part = numpy.zeros( ( 2*nx, np+nx ), dtype=float )

        idmat = numpy.identity( nx, dtype=float )
        ## upper-left  : model partial
        part[:nx,:np] = self.model.partial( xd, pars )
        ## upper-right : model derivative on diagonal
        part[:nx,np:] = self.model.derivative( xd, pars ) * idmat
        ## lower-left  : zeros
        ## lower-right : identity
        part[nx:,np:] = idmat

        return part

    def derivative( self, param ) :
        """
        Return the derivative to the Model.

        Parameters
        ----------
        params : array_like
            list of problem parameters
        """

        ( xd, pars ) = self.splitParam( param )
        return self.model.derivative( xd, pars )

    def domain2Unit( self, dval, kpar ) :
        """
        Return value in [0,1] for the selected parameter.

        Parameters
        ----------
        dval : float
            domain value for the selected parameter
        kpar : int
            selected parameter index, where kp is index in [parameters, hyperparams]
        """
        if kpar < self.model.npars :
            return self.model.domain2Unit( dval, kpar )

        kp = kpar - self.model.npars
        dv = dval - self.xdata[kp]

        return self.prior.domain2Unit( dv )

    def unit2Domain( self, uval, kpar ) :
        """
        Return domain value for the selected parameter.

        Parameters
        ----------
        uval : array_like
            unit value for the selected parameter
        kpar : int
            selected parameter indices, where kp is index in [parameters, hyperparams]
        """
        if kpar < self.model.npars :
            return self.model.unit2Domain( uval, kpar )

        return self.prior.unit2Domain( uval ) + self.xdata[kpar-self.model.npars]

    def getXYresiduals( self, param ) :
        """
        Return residuals in y-direction and x-direction.

        Parameters
        ----------
        param : array_like
            model parameters and xdata parameters

        Returns
        -------
        tuple of (y residuals, x residuals)
        """
        ( xd, pars ) = self.splitParam( param )
        yres = self.ydata - self.model.result( xd, pars )
        return ( yres, self.xdata - xd )

    def weightedResSq( self, allpars, mockdata=None, extra=False ) :
        """
        Return the (weighted) squared distance between (xdata,ydata) and (xtry,ytry) where xtry are
        the trial values for xdata and ytry = model.result( xtry, param )

        Parameters
        ----------
        allpars : array_like
            model parameters, xdata parameters, and noise scale
        mockdata : array_like
            model fit data model.result( xtry, param )
        extra : bool (False)
            true  : return ( wgt * res^2, wgt * [yres,xres] )
            false : return wgt * res^2
        """
        np = self.npars 
        param = allpars[:np] 
        ( yres, xres ) = self.getXYresiduals( param )

        xres2 = xres * xres
        yres2 = yres * yres

        s2 = allpars[np] * allpars[np]
        det = self.determinant + s2 * ( self.varyy + self.varxx + s2 )

        # values from inverse covar matrix
        myy = ( self.varxx + s2 ) / det
        mxx = ( self.varyy + s2 ) / det
        mxy = -self.varxy / det

        yrot2 = yres2 * myy
        xrot2 = xres2 * mxx
        xyrot = xres * yres * mxy

        res2 = xrot2 + yrot2 + 2 * xyrot            # equiprobable distance squared

        if self.weights is not None : 
            res2 *= self.weights

        if not extra :
            return res2 

        xr = xres * mxx + yres * mxy
        yr = yres * myy + xres * mxy

        xyr2 = xres2 + yres2

        if self.weights is not None : 
            xr *= self.weights
            yr *= self.weights
            xyr2 *= self.weights

        return ( res2, numpy.append( yr, xr ), xyr2 )

    def myEngines( self ) :
        """
        Return a default list of preferred engines
        """
        return ["galilean", "gibbs", "chord"]

    def myStartEngine( self ) :
        """
        Return a default preferred start engines: "start"
        """
        return "start"

    def myDistribution( self ) :
        """
        Return a default preferred ErrorDistribution: "gauss2d"
        """
        return "gauss2d"

    #  *****TOSTRING***********************************************************
    def baseName( self ):
        """ Returns a string representation of the model.  """
        return "ErrorsInXandYProblem of %s" % self.model





# run with : python3 -m unittest TestImageAssistant

import unittest
import os
import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import matplotlib.pyplot as plt
import warnings

from BayesicFitting import PolySurfaceModel
from BayesicFitting import ImageAssistant

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestImageAssistant( unittest.TestCase ):
    """
    Test harness for 2d-Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testImageAssistantC( self ):

        print( "====ImageAssistant C ===================" )
        ymap = numpy.arange( 6, dtype=float ).reshape( 2, 3 )
        print( ymap )
        ia = ImageAssistant()
        xdata = ia.getIndices( ymap )
        print( xdata )
        print( xdata.shape )
        for k in range( 6 ) :
            assertAAE( xdata[k,:], [k//3,k%3] )
            self.assertTrue( ymap[xdata[k,0],xdata[k,1]] == k )
        ydata = ia.getydata( ymap )
        print( ydata )
        ydata += 1
        ymap1 = ia.resizeData( ydata )
        print( ymap1 )
        assertAAE( ymap1, ymap + 1 )

        xy0 = ia.getPositions( ymap, center=False )
        print( xy0 )
        self.assertTrue( xy0[0,0] == 0 and xy0[5,1] == 2 )

        xy1 = ia.getPositions( ymap )
        print( xy1 )
        assertAAE( xy1, xy0+0.5 )

        xy2 = ia.getPositions( ymap, deproject=numpy.log )
        print( xy2 )
        assertAAE( xy2, numpy.log( xy1 ) )

        


    def testImageAssistantF( self ):

        print( "====ImageAssistant F ===================" )
        ymap = numpy.arange( 6, dtype=float ).reshape( 2, 3 )
        print( ymap )
        ia = ImageAssistant( order='F')
        xdata = ia.getIndices( ymap )
        print( xdata )
        for k in range( 6 ) :
            assertAAE( xdata[k,:], [k%3,k//3] )
            self.assertTrue( ymap[xdata[k,1],xdata[k,0]] == k )
        ydata = ia.getydata( ymap )
        print( ydata )
        ydata += 1
        ymap1 = ia.resizeData( ydata, shape=ymap.shape )
        print( ymap1 )
        assertAAE( ymap1, ymap + 1 )

    def testImageAssistant3dC( self ):

        print( "====ImageAssistant 3d C ===================" )
        ymap = numpy.arange( 24, dtype=float ).reshape( 2, 3, 4 )
        print( ymap )
        ia = ImageAssistant( order='C')
        xdata = ia.getIndices( ymap )
        print( xdata )
        ydata = ia.getydata( ymap )
        print( ydata )
        ydata += 1
        ymap1 = ia.resizeData( ydata, shape=ymap.shape )
        print( ymap1 )
        assertAAE( ymap1, ymap + 1 )

    def testImageAssistant3dF( self ):

        print( "====ImageAssistant 3d F ===================" )
        ymap = numpy.arange( 24, dtype=float ).reshape( 2, 3, 4 )
        print( ymap )
        ia = ImageAssistant( order='F')
        xdata = ia.getIndices( ymap )
        print( xdata )
        ydata = ia.getydata( ymap )
        print( ydata )
        ydata += 1
        ymap1 = ia.resizeData( ydata )
        print( ymap1 )
        assertAAE( ymap1, ymap + 1 )

    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestImageAssistant.__class__ )

if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
import math

from .ScaledErrorDistribution import ScaledErrorDistribution

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018        Do Kester


class UniformErrorDistribution( ScaledErrorDistribution ):
    """
    To calculate a Uniform likelihood, eg. for digitization noise.

    For one residual, x, it holds

        L( x ) = 1 / ( 2 * s )    if |x| < s
                 0                otherwise

    where s is the scale.
    s is a hyperparameter, which might be estimated from the data.

    The variance of this function is &sigma;^2 = s / 6.
    See: toSigma()

    The function is mostly used to calculate the likelihood L over N residuals,
    or easier using log likelihood, logL.

        logL = -log( 2 * s ) * N

    Note that it is required that <b>all</b> residuals are smaller than s,
    otherwise the logL becomes -inf.

    Using weights this becomes:

        logL = -log( 2 * s ) * &sum; w


    Author       Do Kester.

    """
    SQRT2 = math.sqrt( 2 )
    LGSQ2 = math.log( SQRT2 )
    LOG2 = math.log( 2.0 )


    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, limits=None, copy=None ) :
        """
        Constructor of Uniform Distribution.

        Parameters
        ----------
        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None    no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is *not* fixed.
        copy : UniformErrorDistribution
            distribution to be copied.

        """
        super( UniformErrorDistribution, self ).__init__( scale=scale,
                limits=limits, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return UniformErrorDistribution( copy=self )

    #  *********DATA & WEIGHT***************************************************
    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True

    def getScale( self, problem, allpars=None ) :
        """
        Return the noise scale

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        res = self.getResiduals( problem, allpars=allpars )
        return numpy.max( numpy.abs( res ) )

    def toSigma( self, scale ) :
        """
        Return sigma, the squareroot of the variance.
        Parameter
        --------
        scale : float
            the scale of this Uniform distribution.
        """
        return scale * 2 / math.sqrt( 12 )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ) :
        """
        Return the log( likelihood ) for a Uniform distribution.

        Alternate calculation.

        Outside the range the likelihood is zero, so the logL should be -inf.
        However for computational reasons the maximum negative value is returned.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem

        """
        self.ncalls += 1

        scale = allpars[-1] if not problem.hasAccuracy else problem.accuracy

        ares = numpy.abs( problem.residuals( allpars[:-1] ) )

        if all( ares < scale ) :
            if isinstance( scale, float ) :
                return - numpy.log( 2 * scale ) * problem.sumweight
            elif problem.hasWeights() :
                return - numpy.sum( numpy.log( 2 * scale ) * problem.weights )
            else :
                return - numpy.sum( numpy.log( 2 * scale ) )

        return -math.inf

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual

        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        mockdata : array_like
            as calculated by the model

        """
        if mockdata is None :
            mockdata = problem.result( allpars[:-1] )
        scale = allpars[-1] if not problem.hasAccuracy else problem.accuracy

        ares = numpy.abs( problem.ydata - mockdata )

        lld = numpy.where( ares < scale, -numpy.log( 2 * scale ), -math.inf )
        if problem.weights is not None :
            lld *= problem.weights
        return lld


    def partialLogL_alt( self, problem, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        dL/ds is not implemented for problems with accuracy

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        dL = numpy.zeros( len( fitIndex ), dtype=float )
        if fitIndex[-1] == -1 :
            dL[-1] = -problem.sumweight / allpars[-1]
        return dL

    def nextPartialData( self, problem, allpars, fitIndex, mockdata=None ) :
        """
        Return the partial derivative of elements of the log( likelihood )
        to the parameters.

        dL/ds is not implemented for problems with accuracy

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted
        mockdata : array_like
            as calculated by the model

        """
        scale = allpars[-1]
        pll = numpy.zeros_like( problem.ydata )
        if problem.weights is not None :
            wgt = problem.weights
        else :
            wgt = numpy.ones_like( problem.ydata )

        for k in fitIndex :
            if k >= 0 :
                yield pll
            else :
                yield ( - wgt / scale )

        return

    def __str__( self ) :
        """
        Return a string representation of this class.
        """
        return "Uniform error distribution"




#run with: python3 -m unittest TestCompoundModel

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

#from BayesicFitting import PolynomialModel, SineModel
#from Model import Model

from StdTests import stdModeltest

from BayesicFitting import formatter as fmt
from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testOperations( self ) :
        print( "  Test Operations  " )
        m = GaussModel()
        p = PolynomialModel( 2 )
        m1 = m + p
        self.assertTrue( isinstance( m1, Model ) )
        self.assertTrue( m1.npchain == 6 )

        m = GaussModel()
        p = PolynomialModel( 2 )
        m1 = m - p
        self.assertTrue( isinstance( m1, Model ) )
        self.assertTrue( m1.npchain == 6 )

        m = GaussModel()
        p = PolynomialModel( 2 )
        m1 = m * p
        self.assertTrue( isinstance( m1, Model ) )
        self.assertTrue( m1.npchain == 6 )

        m = GaussModel()
        p = PolynomialModel( 2 )
        m1 = m / p
        self.assertTrue( isinstance( m1, Model ) )
        self.assertTrue( m1.npchain == 6 )

        m = GaussModel()
        p = PolynomialModel( 2 )
        m1 = m | p
        Tools.printclass( m1 )
        self.assertTrue( isinstance( m1, Model ) )
        self.assertTrue( m1.npchain == 6 )


    def testLimits( self ) :
        m = GaussModel()
        m.setLimits( lowLimits=[-10, 5, 0], highLimits=[10, 9, 6] )
        p = PolynomialModel( 3 )
        p.setLimits( lowLimits=[-100], highLimits=[100] )
        m1 = m | p
        print( m1.npchain, m.npchain, p.npchain )
        Tools.printclass( m )
        lo = [m1.getPrior(k).lowLimit for k in range( m1.npchain )]
        print( lo )
        hi = [m1.getPrior(k).highLimit for k in range( m1.npchain )]
        print( hi )

        m1 = p + m

        ( lo, hi ) = m1.getLimits()
        print( "lolim   ", lo )
        print( "hilim   ", hi )
        self.assertTrue( len( lo ) == 7 )
        self.assertTrue( len( hi ) == 7 )


        r = RepeatingModel( 3, m )
        print( r.npchain, m.npchain, m.npbase )
        self.assertTrue( r.npchain == 9 )
        self.assertTrue( r.npbase == 9 )
        self.assertTrue( r.npmax == 9 )
        self.assertTrue( r.ncomp == 3 )
        self.assertTrue( r.deltaNpar == 3 )

        Tools.printclass( r )
        for k in range( r.npchain ) :
            print( k, r.par2model( k ) )
        ( lo, hi ) = r.getLimits()
        print( "lolim   ", lo )
        print( "hilim   ", hi )
        self.assertTrue( len( lo ) == 9 )
        self.assertTrue( len( hi ) == 9 )


        m1 = p + r
        ( lo, hi ) = m1.getLimits()
        print( "lolim   ", lo )
        print( "hilim   ", hi )
        self.assertTrue( len( lo ) == 13 )
        self.assertTrue( len( hi ) == 13 )


        r = RepeatingModel( 3, m, same=2 )
        print( r.npchain, m.npchain, m.npbase )
        self.assertTrue( r.npchain == 7 )
        self.assertTrue( r.npbase == 7 )
        self.assertTrue( r.npmax == 7 )
        self.assertTrue( r.ncomp == 3 )
        self.assertTrue( r.deltaNpar == 2 )
        Tools.printclass( r )
        for k in range( r.npchain ) :
            print( k, r.par2model( k ) )
        lo = [r.getPrior(k).lowLimit for k in range( r.npchain )]
        print( lo )
        m1 = p + r
        ( lo, hi ) = m1.getLimits()
        print( "lolim   ", lo )
        print( "hilim   ", hi )
        self.assertTrue( len( lo ) == 11 )
        self.assertTrue( len( hi ) == 11 )



        m = GaussModel()
        r = RepeatingModel( 3, m, same=2 )
        r.setLimits( lowLimits=[-10, 5, 0], highLimits=[10, 9, 6] )

        print( r.npchain, m.npchain, m.npbase )
        self.assertTrue( r.npchain == 7 )
        self.assertTrue( r.npbase == 7 )
        self.assertTrue( r.npmax == 7 )
        self.assertTrue( r.ncomp == 3 )
        self.assertTrue( r.deltaNpar == 2 )
        Tools.printclass( r )
        for k in range( r.npchain ) :
            print( k, r.par2model( k ) )
        lo = [r.getPrior(k).lowLimit for k in range( r.npchain )]
        print( lo )
        m1 = p + r
        ( lo, hi ) = m1.getLimits()
        print( "lolim   ", lo )
        print( "hilim   ", hi )
        self.assertTrue( len( lo ) == m1.npchain )
        self.assertTrue( len( hi ) == m1.npchain )

        lo = [m1.getPrior(k).lowLimit for k in range( m1.npchain )]
        print( lo )
        hi = [m1.getPrior(k).highLimit for k in range( m1.npchain )]
        print( hi )

        h = HarmonicModel( 3 )
        h.setLimits( lowLimits=[-110,-111], highLimits=[110,111] )
        m1 = p + h
        ( lo, hi ) = m1.getLimits()
        print( "lolim   ", lo )
        print( "hilim   ", hi )
        self.assertTrue( len( lo ) == m1.npchain )
        self.assertTrue( len( hi ) == m1.npchain )

        lo = [m1.getPrior(k).lowLimit for k in range( m1.npchain )]
        print( lo )
        hi = [m1.getPrior(k).highLimit for k in range( m1.npchain )]
        print( hi )

        h = HarmonicDynamicModel( 3 )
        h.setLimits( lowLimits=[-110,-111], highLimits=[110,111] )
        m1 = p + h
        ( lo, hi ) = m1.getLimits()
        print( "lolim   ", lo )
        print( "hilim   ", hi )
        self.assertTrue( len( lo ) == m1.npchain )
        self.assertTrue( len( hi ) == m1.npchain )



    def testOneModel( self ):
        print( "  Test one model" )
        m = GaussModel( )

        self.assertTrue( m.getNumberOfParameters( ) == 3 )
        self.assertTrue( m.npbase == 3 )

        par = numpy.asarray( [3,4,5], dtype=float )
        m.parameters = par

        numpy.testing.assert_array_equal( m.parameters, par )
        self.assertTrue( m.getNumberOfParameters( ) == 3 )
        self.assertTrue( m.npbase == 3 )
        numpy.testing.assert_array_equal( m.parameters, par )

        self.assertTrue( m.getNumberOfParameters( ) == 3 )
        self.assertTrue( m.npbase == 3 )
        par[1] = 10.0
        numpy.testing.assert_array_equal( m.parameters, par )

        fix = numpy.asarray( [0,1,2] )
        self.assertTrue( m.getNumberOfParameters( ) == 3 )
        self.assertTrue( m.npbase == 3 )

    def testTwoModel( self ):
        print( "  Test two models" )

        m = GaussModel( )
        p = PolynomialModel( 1 )

        m.addModel( p )

        print( m.parameters, p.parameters )


        self.assertTrue( m.getNumberOfParameters( ) == 5 )
        self.assertTrue( m.npbase == 3 )
        self.assertTrue( m._next.npbase == 2 )
        self.assertTrue( len( m.parameters ) == 5 )


        par = numpy.asarray( [1,-2,3,-1.2,0.2] )
        m.parameters = par
        print( m.parameters )
        numpy.testing.assert_array_equal( m.parameters, par )

    def testThreeModel( self ):
        print( "  Test three models" )
        m = GaussModel( )
        self.assertTrue( m.chainLength() == 1 )
        p = PolynomialModel( 1 )
        m.addModel( p )
        self.assertTrue( m.chainLength() == 2 )
        s = VoigtModel( )
        m.addModel( s )
        self.assertTrue( m.chainLength() == 3 )

        params = numpy.arange( 9, dtype=float )
        m.parameters = params

        self.assertTrue( m._next == p )
        self.assertTrue( m._next._next == s )
        self.assertTrue( p._next == s )
        self.assertTrue( s._head == m )
        self.assertTrue( p._head == m )
        self.assertTrue( m._head == m )
        self.assertTrue( m.getNumberOfParameters( ) == 9 )
        self.assertTrue( p.getNumberOfParameters( ) == 9 )
        self.assertTrue( s.getNumberOfParameters( ) == 9 )
        self.assertTrue( m.npbase == 3 )
        self.assertTrue( p.npbase == 2 )
        self.assertTrue( s.npbase == 4 )
        self.assertTrue( len( m.parameters ) == 9  )
        self.assertTrue( p.parameters is None  )
        self.assertTrue( s.parameters is None  )


        print( "Isolate second model into p1" )
        p1 = m.isolateModel( 1 )
        print( p1 )
        print( p1.parameters )
        self.assertTrue( isinstance( p1, PolynomialModel ) )

        self.assertTrue( p1._next == None )
        self.assertTrue( p1._head == p1 )
        self.assertTrue( p1.npbase == 2 )
        self.assertTrue( p1.getNumberOfParameters() == 2 )


        self.assertTrue( m._next == p )
        self.assertTrue( m._next._next == s )
        self.assertTrue( p._next == s )
        self.assertTrue( s._head == m )
        self.assertTrue( p._head == m )
        self.assertTrue( m._head == m )
        self.assertTrue( m.getNumberOfParameters( ) == 9 )
        self.assertTrue( p.getNumberOfParameters( ) == 9 )
        self.assertTrue( s.getNumberOfParameters( ) == 9 )
        self.assertTrue( m.npbase == 3 )
        self.assertTrue( p.npbase == 2 )
        self.assertTrue( s.npbase == 4 )


    def testPipe1( self ) :
        model = PolynomialModel( 2 )
        model |= SineModel()
        print( model )
        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 1.0] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        y = model.result( x, p )

        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe2( self ) :
        model = PolynomialModel( 2 )
        model |= SineModel()
        model *= ExpModel()
        print( model )
        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 1.0, 1.3, -0.2] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        y = model.result( x, p )

        stdModeltest( model, p, x=x, plot=self.doplot )

    def testPipe3( self ) :
        model = StellarOrbitModel( )
        model |= PolySurfaceModel( 3 )

        print( model.npars )
        printclass( model )
        printclass( model._next )

        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4, 0.2, 1.0, 0.0, 0.0, 0.4,
                            -0.1, -0.2, 0.3, 0.0, 0.0] )
        x = numpy.linspace( 0, 10, 101, dtype=float )

        y = model.result( x, p )
        print( y.shape )
        print( fmt( y ) )
        pp = model.partial( x, p )
        print( pp.shape )
        print( fmt( pp ) )
        dd = model.derivative( x, p )
        print( dd.shape )
        print( fmt( dd ) )

        stdModeltest( model, p, x=x, plot=self.doplot )


    def testThreeModelLimits( self ):
        print( "  Test three models: limits and domain <> unit" )

        m = GaussModel( )
        p = PolynomialModel( 1 )
        m.addModel( p )
        s = VoigtModel( )
        m.addModel( s )

        print( m.npchain )

        lo = numpy.zeros( 9 )
        hi = numpy.arange( 9 ) + 10.0
        m.setLimits( lo, hi )

        par = numpy.arange( 9, dtype=float )
        unp = m.domain2Unit( par )
        print( fmt( par, max=None ) )
        print( fmt( lo, max=None ) )
        print( fmt( hi, max=None ) )
        print( fmt( unp, max=None ) )
        print( fmt( par / hi, max=None ) )
        print( fmt( m.unit2Domain( unp ), max=None ) )
        print( fmt( m.partialDomain2Unit( par ), max=None ) )

        numpy.testing.assert_array_almost_equal( m.unit2Domain( unp ), par, 8 )
        self.assertTrue( m.domain2Unit( par[2], 2 ) == unp[2] )
        self.assertTrue( m.domain2Unit( par[3], 3 ) == unp[3] )
        self.assertTrue( m.domain2Unit( par[6], 6 ) == unp[6] )

    def testTwoModelDerivatives( self ):
        print( "  Test two model derivatives" )
        m = GaussModel( )
        m.addModel( PolynomialModel(1) )
        p = numpy.asarray( [1,-2,3,-1.2,0.2] )
        x = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )

        print( m )
        m.testPartial( x[3], p )

        part = m.partial( x, p )
        nump = m.numPartial( x, p )
        numpy.testing.assert_array_almost_equal( part, nump, 4 )

        mc = m.copy( )
        self.assertTrue( mc.getNumberOfParameters( ) == 5 )
        self.assertTrue( mc.getNumberOfFittedParameters( ) == 5 )

        mc.parameters = p
        m.parameters = p

        numpy.testing.assert_array_equal( m.result( x ), mc.result( x ) )

        m.addModel( GaussModel() )
        m.parameters =  numpy.asarray( [1,-2,3,1,-2,2,2,3], dtype=float )
        self.assertTrue( m.parameters[2] > 0 )
        self.assertTrue( m.parameters[7] > 0 )
        m.parameters = numpy.asarray( [1,-2,-3,1,-2,2,2,-3], dtype=float )
        m.result( x )
        self.assertTrue( m.parameters[2] > 0 )
        self.assertTrue( m.parameters[7] > 0 )
        m.parameters = numpy.asarray( [1,-2,0,1,-2,2,2,0], dtype=float )
        self.assertWarns( UserWarning, m.result, x )
        self.assertTrue( m.parameters[2] == m.tiny )
        self.assertTrue( m.parameters[7] == m.tiny )
        m.result( x )

    def testTwoModelSubtractDerivatives( self ):
        print( "  Test two model subtract derivatives" )
        m = GaussModel( )
        m.subtractModel( PolynomialModel(1) )
        p = numpy.asarray( [1,-2,3,-1.2,0.2] )
        x = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )

        print( m )
        m.testPartial( x[3], p )

        part = m.partial( x, p )
        nump = m.numPartial( x, p )
        numpy.testing.assert_array_almost_equal( part, nump, 4 )

        mc = m.copy( )
        self.assertTrue( mc.getNumberOfParameters( ) == 5 )

        mc.parameters = p
        m.parameters = p

        numpy.testing.assert_array_equal( m.result( x ), mc.result( x ) )

    def testTwoModelDerivatives( self ):
        print( "  Test two model multiply derivatives" )
        m = GaussModel( )
        m.multiplyModel( PolynomialModel(1) )
        p = numpy.asarray( [1,-2,3,-1.2,0.2] )
        x = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )

        print( m )
        m.testPartial( x[3], p )

        part = m.partial( x, p )
        nump = m.numPartial( x, p )
        numpy.testing.assert_array_almost_equal( part, nump, 4 )

        mc = m.copy( )
        self.assertTrue( mc.getNumberOfParameters( ) == 5 )

        mc.parameters = p
        m.parameters = p

        numpy.testing.assert_array_equal( m.result( x ), mc.result( x ) )

    def testTwoModelDivideDerivatives( self ):
        print( "  Test two model divide derivatives" )
        m = GaussModel( )
        m.divideModel( PolynomialModel(1) )
        p = numpy.asarray( [1,-2,3,-1.2,0.2] )
        x = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )

        print( m )
        m.testPartial( x[3], p )

        part = m.partial( x, p )
        nump = m.numPartial( x, p )
        numpy.testing.assert_array_almost_equal( part, nump, 4 )

        mc = m.copy( )
        self.assertTrue( mc.getNumberOfParameters( ) == 5 )

        mc.parameters = p
        m.parameters = p

        numpy.testing.assert_array_equal( m.result( x ), mc.result( x ) )

    def testModelName( self ):
        print( "  Test model names" )

        pm = PolynomialModel( 2 )
        gm = GaussModel( fixed={0:pm} )
        sm = SineModel()
        m1 = gm / sm
        self.assertTrue( gm.npbase == 5 )
        self.assertTrue( m1.npchain == 8 )
        print( '-------------------------------------------------------------' )
        print( m1 )

        print( '-------------------------------------------------------------' )
        print( m1._toString( "toS   ", npars=2 ) )

        sm = SplinesModel( [0,1,2] )
        pm = PolynomialModel( 0 )
        lm = LorentzModel( )
        m3 = sm + lm + pm
        print( '-------------------------------------------------------------' )
        print( m3 )




        em = PowerModel( 1, fixed={0:1.0} )
        em += BasicSplinesModel( nrknots=8, min=0, max=10 )
        em += ExpModel()
        m2 = em | m1
        self.assertTrue( em.npbase == 0 )
        self.assertTrue( em.npchain == 12 )
        self.assertTrue( m2.npchain == 20 )
        print( '-------------------------------------------------------------' )
        m2str = m2.__str__()
        print( m2str )

        self.assertTrue( m2str[-30:] == "p_19 * sin( 2PI * x * p_17 ) }" )

        for k in range( m2.npchain ) :
            print( fmt( k ), "%-13.13s"%m2.getParameterName(k), fmt( m2[k] ) )







    def suite( cls ):
        return unittest.TestCase.suite( CompoundModelTest.__class__ )


if __name__ == '__main__':
    unittest.main()




import numpy as numpy
from . import Tools
from .Tools import setAttribute as setatt

from .Model import Model
from .PolynomialModel import PolynomialModel
from .PolySurfaceModel import PolySurfaceModel

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2011 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2022 Do Kester

class ConstantModel( Model ):
    """
    ConstantModel is a Model which does not have any parameters.

        f( x:p ) = f( x )

    As such it is irrelevant whether it is linear or not.
    It has 0 params and returns a 0 for its partials.

    ConstantModel, by default, returns a constant ( = 0 ) for its result.
    It can however return any fixed form that a Model can provide.

    This might all seem quite irrelevant for fitting. And indeed no parameters
    can be fitted to these models, no standard deviations can be calculated, but
    it is possible to calculate the evidence for these models and compare them
    with more complicated models to decide whether there is any evidence for
    some structure at all.

    It can also be used when some constant is needed in a compound model,
    or a family of similar shapes.

    Attributes
    ----------
    fixedModel : Model
        a model which is calculated. (default: 0, everywhere)
    table : array_like
        array of tabulated results

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    Examples
    --------
    To make a model that decays to 1.0

    >>> model = ConstantModel( values=1.0 )
    >>> model.addModel( ExpModel( ) )

    To make a model that returns a fixed cosine of frequency 5

    >>> model = ConstantModel( fixedModel=SineModel(), values=[1.0,0.0,5.0] )

    """
    def __init__( self, ndim=1, copy=None, fixedModel=None, values=None,
                    table=None ):
        """
        The ConstantModel implementation.

        Number of parameters = 0.

        Parameters
        ----------
        ndim : int
            number of dimensions for the model. (default: 1)
        copy : ConstantModel
            model to be copied. (default: None)
        fixedModel : Model
            a fixed model to be returned. (default: 0 everywhere)
        values : array_like
            parameters to be used in the fixedModel. (default: None)
        table : array_like
            array of tabulated results

        Notes
        -----
        A table provided to the constructor has only values at the xdata.
        At other vales than xdata, the model does not work.

        """
        super( ConstantModel, self ).__init__( 0, ndim=ndim, copy=copy )

        if copy is not None :                  # copy from model
            setatt( self, "fixedModel", copy.fixedModel )
            setatt( self, "table", copy.table )
        elif table is not None :                # store tabular results
            setatt( self, "table", table, islist=True, type=float )
            setatt( self, "fixedModel", None )
        else :                                  # make a fixedModel
            setatt( self, "table", None )
            if fixedModel is not None :
                setatt( self, "fixedModel", fixedModel, type=Model )
            else :
                setatt( self, "fixedModel", PolynomialModel( 0 ) if ndim == 1 else PolySurfaceModel( 0 ) )
            if values is not None :
                setatt( self.fixedModel, "parameters", values, type=float, islist=True )

    def copy( self ):
        """ Copy method.  """
        return ConstantModel( ndim=self.ndim, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: fixedModel, table, values
        """
        if name == "fixedModel" :
            setatt( self, name, value, type=Model, isnone=True )
            return
        if name == "table" :
            setatt( self, name, value, type=float, islist=True, isnone=True )
            setatt( self, "fixedModel", None )
            return
        if name == "values" or name == "value":
            setatt( self.fixedModel, "parameters", value, type=float, islist=True )
            return

        super( ConstantModel, self ).__setattr__( name, value )

    def baseResult( self, xdata, params ):
        """
        Returns a constant form.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters. (irrelevant)

        """
        if self.table is not None :
            return self.table

        return self.fixedModel.result( xdata )

    def basePartial( self, xdata, params ) :
        """
        Returns the partials at the xdata value. (=empty array)

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters. (irrelevant)

        """
        return numpy.zeros( ( Tools.length( xdata ), self.npbase ), dtype=float )

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each point x (== 0).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters. (irrelevant)

        """
        if self.fixedModel is None :
            return numpy.zeros_like( xdata )

        return self.fixedModel.derivative( xdata, self.fixedModel.parameters )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        content = "tabular result" if self.table is not None else self.fixedModel.shortName()
        return str( "ConstantModel: f( x ) = " + content )


